import sys

from ...utils import indent

__id_node = 0
def xml_representation(fsm, tlacenode, spec):
    """
    Return the XML representation of tlacenode explaining spec violation or
    satisfaction by fsm.
    
    Return the XML representation of a TLACE
    starting at tlacenode, explaining why the state of tlacenode,
    belonging to fsm, violates or satisfies spec.
    
    fsm -- the FSM violating spec.
    tlacenode -- the TLACE node explaining the violation or satisfaction
                 of spec by fsm.
    spec -- the violated specification.
    """
    
    indent.reset()
    
    global __id_node
    __id_node = 0
    
    # Open counterexample
    xmlrepr = (
    """<?xml version="1.0" encoding="UTF-8"?>
<tlace specification="{spec}">
""".format(spec=str(spec)))
    
    indent.inc()
    
    xmlrepr += xml_node(fsm, tlacenode)
    
    indent.dec()
    
    xmlrepr += """</tlace>"""
    
    return xmlrepr
    
    
def xml_node(fsm, node):
    """
    Return the XML representation of the given TLACE node.
    
    fsm -- the FSM of the node.
    node -- the TLACE node to represent.
    """
    
    # node tag
    global __id_node
    
    xmlrepr = indent.indent(
    """<node id="{0}">
""".format(__id_node))
    
    __id_node += 1
    
    indent.inc()
    # state node
    xmlrepr += xml_state(fsm, node.state)
    
    # atomics
    for atomic in node.atomics:
        xmlrepr += indent.indent(
        """<atomic specification="{0}" />
""".format(str(atomic)))
    
    # branches
    for branch in node.branches:
        xmlrepr += xml_branch(fsm, branch)
    
    # universals
    for universal in node.universals:
        xmlrepr += indent.indent(
        """<universal specification="{0}" />
""".format(str(universal)))
    
    indent.dec()
    
    xmlrepr += indent.indent("""</node>
""")
    
    return xmlrepr
    
    
def xml_branch(fsm, branch):
    """
    Return the XML representation of the given TLACE branch.
    
    fsm -- the FSM of the node.
    branch -- the TLACE branch to represent.
    """
    
    loop_id = -1
    
    xmlrepr = indent.indent(
    """<existential specification="{0}">
""".format(str(branch.specification)))
    
    indent.inc()
    
    for n, i in zip(branch.path[0][::2], branch.path[0][1::2]):
        xmlrepr += xml_node(fsm, n)
        xmlrepr += xml_inputs(fsm, i)
        xmlrepr += xml_combinatorial(fsm, i)
        if branch.path[1] is not None and n == branch.path[1][1]:
            loop_id = __id_node - 1
        
    xmlrepr += xml_node(fsm, branch.path[0][-1])
    
    if branch.path[1] is not None:
        xmlrepr += xml_inputs(fsm, branch.path[1][0])
        xmlrepr += indent.indent(
        """<loop to="{0}" />
""".format(loop_id))
    
    indent.dec()
    xmlrepr += indent.indent(
    """</existential>
""")
    
    return xmlrepr
    
    
def xml_state(fsm, state):
    """
    Return the XML representation of the given state.
    
    fsm -- the FSM of the state.
    state -- a BDD representing a state of fsm.
    """
    
    xmlrepr = indent.indent(
    """<state>
""")
    
    indent.inc()
    
    values = state.get_str_values()
    for var in values:
        xmlrepr += indent.indent(
        """<value variable="{0}">{1}</value>
""".format(var, values[var]))
    
    indent.dec()
    xmlrepr += indent.indent("""</state>
""")
    
    return xmlrepr
    

def xml_inputs(fsm, inputs):
    """
    Return the XML representation of the given inputs.
    
    fsm -- the FSM.
    state -- a BDD representing inputs in fsm.
    """
        
    xmlrepr = indent.indent(
    """<input>
""")
    
    indent.inc()
    
    values = inputs.get_str_values()
    for var in values:
        xmlrepr += indent.indent(
        """<value variable="{0}">{1}</value>
""".format(var, values[var]))
    
    indent.dec()
    xmlrepr += indent.indent("""</input>
""")
    
    return xmlrepr
    
    
def xml_combinatorial(fsm, combinatorial):
    """
    Return the XML representation of the given combinatorial.
    
    fsm -- the FSM.
    state -- a BDD representing inputs in fsm.
    """
        
    xmlrepr = indent.indent(
    """<combinatorial>
""")
    
    indent.inc()
    
    # TODO Find a way to get combinatorial
    
    indent.dec()
    xmlrepr += indent.indent("""</combinatorial>
""")
    
    return xmlrepr


from pynusmv.dd import BDD

from .eval import ceg, ceu, cew
from pynusmv_tools.explanation.explanation import Explanation

def explain_cex(fsm, state, agents, phi):
    """
    Explain why state of fsm satisfies <agents> X phi.
    
    fsm -- the fsm;
    state -- a state of fsm satisfying <agents> X phi;
    agents -- the agents of the specification;
    phi -- the set of states of fsm satifying phi.
    """
    # To show that state satisfies <agents> X phi, we have to exhibit an action
    # of agents such that all reached states are in phi.
    # We choose to execute this action, that is, exhibit the state, and
    # all the reached states through the action.
    
    # Cubes
    gamma_cube = fsm.inputs_cube_for_agents(agents)
    ngamma_cube = fsm.bddEnc.inputsCube - gamma_cube
    
    # Get the winning actions, for all state
    si = fsm.pre_strat_si(phi, agents)
    # The winning actions from state are the ones belonging to its protocol
    # and to si
    actions = (si & (fsm.protocol(agents) & state) &
               fsm.bddEnc.inputsMask).forsome(fsm.bddEnc.statesCube)
    fullaction = fsm.pick_one_inputs(actions)
    ag_action = fullaction.forsome(ngamma_cube)
    
    
    # Build the graph
    successors = set()
    for successor in fsm.pick_all_states(fsm.post(state, ag_action) &
                                         fsm.bddEnc.statesMask):
        nextexpl = Explanation(successor)
        allactions = fsm.get_inputs_between_states(state, successor) & ag_action 
        for action in fsm.pick_all_inputs(allactions & fsm.bddEnc.inputsMask):
            successors.add((action, nextexpl))
    return Explanation(state, successors)
    
    
def explain_ceu(fsm, state, agents, phi, psi, states=None):
    """
    Explain why state of fsm satisfies <agents> phi U psi.
    
    fsm -- the fsm;
    state -- a state of fsm satisfying <agents> phi U psi;
    agents -- the agents of the specification;
    phi -- the set of states of fsm satifying phi;
    psi -- the set of states of fsm satifying psi;
    states -- a dictionary of state->explanation pairs.
    """
    # <agents> phi U psi = mu Y. psi | (phi & fsm.pre_strat(Y, agents))
    
    # To show that state satisfies <agents> phi U psi
    #   if state in psi, return Explanation(state) because state is its own
    #   explanation
    #   otherwise, compute the fixpoint, until state is in the last set
    #   then choose one action always leading to states of the previous set
    #   (it is possible because state is added because it can go one step in
    #   the right direction), and recursively explain why the found states
    #   satisfy <agents> phi U psi.
    
    # states is used to store already explained states and create a directed
    # acyclic graph instead of the pure execution tree, to avoid a blow up
    # in terms of number of states.
    
    if states is None:
        states = {}
    
    if state in states:
        return states[state]
    
    if state <= psi:
        new = Explanation(state)
        states[state] = new
        return new
        
    else:
        f = lambda Y : psi | (phi & fsm.pre_strat(Y, agents))
        old = BDD.false(fsm.bddEnc.DDmanager)
        new = psi
        while not state <= new:
            old = new
            new = f(old)
        
        expl = explain_cex(fsm, state, agents, old)
        
        successors = set()
        for action, succ in expl.successors:
            successors.add((action,
                            explain_ceu(fsm, succ.state, agents,
                                        phi, psi, states)))
        new = Explanation(state, successors)
        states[state] = new
        return new
    
    
def explain_ceg(fsm, state, agents, phi):
    """
    Explain why state of fsm satisfies <agents> G phi.
    
    fsm -- the fsm;
    state -- a state of fsm satisfying <agents> G phi;
    agents -- the agents of the specification;
    phi -- the set of states of fsm satifying phi.
    """
    # To show that state satisfies <agents> G phi
    # we just have to explain why state satisfies <agents>X<agents>G phi
    # and iterate until we reach the fixpoint.
    # By keeping track of already visited states,
    # we know that we will finally extract the full sub-system of phi states.
    
    # Get CEG(phi)
    cegphi = ceg(fsm, agents, phi)
    
    # Get all states and transitions
    extract = {state}
    states = set()
    transitions = set()
    while len(extract) > 0:
        cur = extract.pop()
        states.add(cur)
        expl = explain_cex(fsm, cur, agents, cegphi)
        for act, succ in expl.successors:
            transitions.add((cur, act, succ.state))
            if succ.state not in states:
                extract.add(succ.state)
    
    # Build the graph
    nodes = {}
    for s in states:
        nodes[s] = Explanation(s)
    for cur, act, succ in transitions:
        nodes[cur].successors.add((act, nodes[succ]))
    
    return nodes[state]
    
    
def explain_cew(fsm, state, agents, phi, psi):
    """
    Explain why state of fsm satisfies <agents> phi W psi.
    
    fsm -- the fsm;
    state -- a state of fsm satisfying <agents> phi W psi;
    agents -- the agents of the specification;
    phi -- the set of states of fsm satifying phi;
    psi -- the set of states of fsm satifying psi.
    """
    # To show that state satisfies <agents> phi W psi
    # either it satisfies psi and we are done,
    # or we have to explain why state satisfies <agents>X<agents>phi W psi
    # and iterate until we reach the fixpoint or stop at psi.
    # By keeping track of already visited states,
    # we know that we will finally extract the full sub-system.
    
    # Get CEW(phi,psi)
    cewpp = cew(fsm, agents, phi, psi)
    
    # Get all states and transitions
    extract = {state}
    states = set()
    transitions = set()
    while len(extract) > 0:
        cur = extract.pop()
        states.add(cur)
        if not cur <= psi:
            expl = explain_cex(fsm, cur, agents, cewpp)
            for act, succ in expl.successors:
                transitions.add((cur, act, succ.state))
                if succ.state not in states:
                    extract.add(succ.state)
    
    # Build the graph
    nodes = {}
    for s in states:
        nodes[s] = Explanation(s)
    for cur, act, succ in transitions:
        nodes[cur].successors.add((act, nodes[succ]))
    
    return nodes[state]
    
    
def explain_cax(fsm, state, agents, phi):
    """
    Explain why state of fsm satisfies [agents] X phi.
    
    fsm -- the fsm;
    state -- a state of fsm satisfying [agents] X phi;
    agents -- the agents of the specification;
    phi -- the set of states of fsm satifying phi.
    """
    # To show that state satisfies [agents] X phi, we have to show that
    # for each action of agents, there is a completion by other agents such that
    # the reached state satisfies phi.
    # That is, we have to take all actions and of agents and for each of them
    # exhibit a successor satisfying phi.
    
    # Cubes
    gamma_cube = fsm.inputs_cube_for_agents(agents)
    ngamma_cube = fsm.bddEnc.inputsCube - gamma_cube
    
    # Get all different actions of agents
    ag_actions = set()
    for action in fsm.pick_all_inputs((fsm.protocol(agents) &
                                         state).forsome(fsm.bddEnc.statesCube)):
        ag_actions.add(action.forsome(ngamma_cube))
    
    # Get successors
    successors = set()
    for ag_action in ag_actions:
        # Choose one successor through ag_action
        succ = fsm.pick_one_state(fsm.post(state, ag_action) & phi &
                                  fsm.bddEnc.statesMask)
        # Get the full action
        act = fsm.pick_one_inputs(fsm.get_inputs_between_states(state, succ) &
                                  ag_action & fsm.bddEnc.inputsMask)
        successors.add((act, Explanation(succ)))
    
    return Explanation(state, successors)    
    
    
def explain_cau(fsm, state, agents, phi, psi, states=None):
    """
    Explain why state of fsm satisfies [agents] phi U psi.
    
    fsm -- the fsm;
    state -- a state of fsm satisfying [agents] phi U psi;
    agents -- the agents of the specification;
    phi -- the set of states of fsm satifying phi;
    psi -- the set of states of fsm satifying psi;
    states -- a dictionary of state->explanation pairs.
    """
    # [agents] phi U psi = mu Y. psi | (phi & fsm.pre_nstrat(Y, agents))
    
    # To show that state satisfies [agents] phi U psi
    #   if state in psi, return Explanation(state) because state is its own
    #   explanation
    #   otherwise, compute the fixpoint, until state is in the last set
    #   then show that all actions always lead to states of the previous set
    #   (it is possible because state is added because it cannot avoid to go
    #   one step in the right direction), and recursively explain why the found
    #   states satisfy [agents] phi U psi.
    
    # states is used to store already explained states to create a directed
    # acyclic graph instead of the pure execution tree.
    
    if states is None:
        states = {}
        
    if state in states:
        return states[state]
    
    if state <= psi:
        new = Explanation(state)
        states[state] = new
        return new
        
    else:
        f = lambda Y : psi | (phi & fsm.pre_nstrat(Y, agents))
        old = BDD.false(fsm.bddEnc.DDmanager)
        new = psi
        while not state <= new:
            old = new
            new = f(old)
        
        expl = explain_cax(fsm, state, agents, old)
        
        successors = set()
        for action, succ in expl.successors:
            successors.add((action,
                            explain_cau(fsm, succ.state, agents,
                                        phi, psi, states)))
        new = Explanation(state, successors)
        states[state] = new
        return new
    
    
def explain_cag(fsm, state, agents, phi):
    """
    Explain why state of fsm satisfies [agents] G phi.
    
    fsm -- the fsm;
    state -- a state of fsm satisfying [agents] G phi;
    agents -- the agents of the specification;
    phi -- the set of states of fsm satifying phi.
    """
    # To show that state satisfies [agents] G phi
    # we just have to explain why state satisfies [agents]X[agents]G phi
    # and iterate until we reach the fixpoint.
    # By keeping track of already visited states,
    # we know that we will finally extract the full sub-system of phi states.
    
    # Get CAG(phi)
    cagphi = ~ceu(fsm, agents, BDD.true(fsm.bddEnc.DDmanager), ~phi)
    
    # Get all states and transitions
    extract = {state}
    states = set()
    transitions = set()
    while len(extract) > 0:
        cur = extract.pop()
        states.add(cur)
        expl = explain_cax(fsm, cur, agents, cagphi)
        for act, succ in expl.successors:
            transitions.add((cur, act, succ.state))
            if succ.state not in states:
                extract.add(succ.state)
    
    # Build the graph
    nodes = {}
    for s in states:
        nodes[s] = Explanation(s)
    for cur, act, succ in transitions:
        nodes[cur].successors.add((act, nodes[succ]))
    
    return nodes[state]
    
    
def explain_caw(fsm, state, agents, phi, psi):
    """
    Explain why state of fsm satisfies [agents] phi W psi.
    
    fsm -- the fsm;
    state -- a state of fsm satisfying [agents] phi W psi;
    agents -- the agents of the specification;
    phi -- the set of states of fsm satifying phi;
    psi -- the set of states of fsm satifying psi.
    """
    # To show that state satisfies [agents] phi W psi
    # either it satisfies psi and we are done,
    # or we have to explain why state satisfies [agents]X[agents]phi W psi
    # and iterate until we reach the fixpoint or stop at psi.
    # By keeping track of already visited states,
    # we know that we will finally extract the full sub-system.
    
    # Get CEW(phi,psi)
    cawpp = ~ceu(fsm, agents, ~psi, ~psi & ~phi)
    
    # Get all states and transitions
    extract = {state}
    states = set()
    transitions = set()
    while len(extract) > 0:
        cur = extract.pop()
        states.add(cur)
        if not cur <= psi:
            expl = explain_cax(fsm, cur, agents, cawpp)
            for act, succ in expl.successors:
                transitions.add((cur, act, succ.state))
                if succ.state not in states:
                    extract.add(succ.state)
    
    # Build the graph
    nodes = {}
    for s in states:
        nodes[s] = Explanation(s)
    for cur, act, succ in transitions:
        nodes[cur].successors.add((act, nodes[succ]))
    
    return nodes[state]
    
    
if __name__ == '__main__':
    import argparse, sys
    from pynusmv.init import init_nusmv, deinit_nusmv
    from ..mas import glob
    from .parsing import parseATL
    from .eval import evalATL
    from .check import check
    from pyparsing import ParseException
    from pynusmv.exception import PyNuSMVError
    from .ast import CEF, CEG, CEX, CEU, CEW, CAF, CAG, CAX, CAU, CAW
    
    def explain_top(mas, spec):
        """
        Explain why the top operator of spec is satisfied.
        
        mas -- the concerned MAS;
        spec -- the given specification; must be satisfied by mas.
        """
        
        # Get satisfying initial state
        sat = evalATL(mas, spec)
        state = mas.pick_one_state(
                                 (sat & mas.bddEnc.statesInputsMask & mas.init))
                                
        # Switch between operators
        if type(spec) is CEF:
            phi = BDD.true(mas.bddEnc.DDmanager)
            psi = evalATL(mas, spec.child)
            agents = {atom.value for atom in spec.group}
            return explain_ceu(mas, state, agents, phi, psi)
            
        elif type(spec) is CEG:
            phi = evalATL(mas, spec.child)
            agents = {atom.value for atom in spec.group}
            return explain_ceg(mas, state, agents, phi)
            
        elif type(spec) is CEX:
            phi = evalATL(mas, spec.child)
            agents = {atom.value for atom in spec.group}
            return explain_cex(mas, state, agents, phi)
            
        elif type(spec) is CEU:
            phi = evalATL(mas, spec.left)
            psi = evalATL(mas, spec.right)
            agents = {atom.value for atom in spec.group}
            return explain_ceu(mas, state, agents, phi, psi)
            
        elif type(spec) is CEW:
            phi = evalATL(mas, spec.left)
            psi = evalATL(mas, spec.right)
            agents = {atom.value for atom in spec.group}
            return explain_cew(mas, state, agents, phi, psi)
            
        elif type(spec) is CAF:
            phi = BDD.true(mas.bddEnc.DDmanager)
            psi = evalATL(mas, spec.child)
            agents = {atom.value for atom in spec.group}
            return explain_cau(mas, state, agents, phi, psi)
            
        elif type(spec) is CAG:
            phi = evalATL(mas, spec.child)
            agents = {atom.value for atom in spec.group}
            return explain_cag(mas, state, agents, phi)
            
        elif type(spec) is CAX:
            phi = evalATL(mas, spec.child)
            agents = {atom.value for atom in spec.group}
            return explain_cax(mas, state, agents, phi)
            
        elif type(spec) is CAU:
            phi = evalATL(mas, spec.left)
            psi = evalATL(mas, spec.right)
            agents = {atom.value for atom in spec.group}
            return explain_cau(mas, state, agents, phi, psi)
            
        elif type(spec) is CAW:
            phi = evalATL(mas, spec.left)
            psi = evalATL(mas, spec.right)
            agents = {atom.value for atom in spec.group}
            return explain_caw(mas, state, agents, phi, psi)
            
        else:
            return Explanation(state)
            
        
    def cntex_top(mas, spec):
        """
        Explain why the top operator of spec is violated.
        
        mas -- the concerned MAS;
        spec -- the given specification; must be violated by mas.
        """
        
        # Get violating initial state
        sat = evalATL(mas, spec)
        state = mas.pick_one_state(
                                (~sat & mas.bddEnc.statesInputsMask & mas.init))
                                
        # Switch between operators
        if type(spec) is CEF:
            # ~<a> F p = [a] G ~p
            phi = ~evalATL(mas, spec.child)
            agents = {atom.value for atom in spec.group}
            return explain_cag(mas, state, agents, phi)
            
        elif type(spec) is CEG:
            # ~<a> G p = [a] F ~p = [a] true U ~p
            phi = BDD.true(mas.bddEnc.DDmanager)
            psi = ~evalATL(mas, spec.child)
            agents = {atom.value for atom in spec.group}
            return explain_cau(mas, state, agents, phi, psi)
            
        elif type(spec) is CEX:
            # ~<a> X p = [a] X ~p
            phi = ~evalATL(mas, spec.child)
            agents = {atom.value for atom in spec.group}
            return explain_cax(mas, state, agents, phi)
            
        elif type(spec) is CEU:
            # ~<a> p U q = [q] ~q W ~p & ~q
            p = evalATL(mas, spec.left)
            q = evalATL(mas, spec.right)
            phi = ~q
            psi = ~p & ~q
            agents = {atom.value for atom in spec.group}
            return explain_caw(mas, state, agents, phi, psi)
            
        elif type(spec) is CEW:
            # ~<a> p W q = [q] ~q U ~p & ~q
            p = evalATL(mas, spec.left)
            q = evalATL(mas, spec.right)
            phi = ~q
            psi = ~p & ~q
            agents = {atom.value for atom in spec.group}
            return explain_cau(mas, state, agents, phi, psi)
            
        elif type(spec) is CAF:
            # ~[a] F p = <a> G ~p
            phi = ~evalATL(mas, spec.child)
            agents = {atom.value for atom in spec.group}
            return explain_ceg(mas, state, agents, phi)
            
        elif type(spec) is CAG:
            # ~[a] G p = <a> F ~p = <a> true U ~p
            phi = BDD.true(mas.bddEnc.DDmanager)
            psi = ~evalATL(mas, spec.child)
            agents = {atom.value for atom in spec.group}
            return explain_ceu(mas, state, agents, phi, psi)
            
        elif type(spec) is CAX:
            # ~[a] X p = <a> X ~p
            phi = ~evalATL(mas, spec.child)
            agents = {atom.value for atom in spec.group}
            return explain_cex(mas, state, agents, phi)
            
        elif type(spec) is CAU:
            # ~[a] p U q = <q> ~q W ~p & ~q
            p = evalATL(mas, spec.left)
            q = evalATL(mas, spec.right)
            phi = ~q
            psi = ~p & ~q
            agents = {atom.value for atom in spec.group}
            return explain_cew(mas, state, agents, phi, psi)
            
        elif type(spec) is CAW:
            # ~[a] p W q = <q> ~q U ~p & ~q
            p = evalATL(mas, spec.left)
            q = evalATL(mas, spec.right)
            phi = ~q
            psi = ~p & ~q
            agents = {atom.value for atom in spec.group}
            return explain_ceu(mas, state, agents, phi, psi)
            
        else:
            return Explanation(state)
    
    
    # Parse arguments
    description = """ATL model checking explainer.
    
    The truth value of the given property is printed on stderr,
    a dot-formated explanation is printed on stdout.
    A counter-example is printed if the property is violated,
    a witness is printed otherwise.
    Only the top operator is explained.
    """
    parser = argparse.ArgumentParser(description=description)
    # Populate arguments: the model, the property
    parser.add_argument('model', help='the MAS as an SMV model')
    parser.add_argument('property', help='the property to explain')
    args = parser.parse_args()
    
    init_nusmv()
    # Initialize the model
    glob.load_from_file(args.model)
    mas = glob.mas()
    
    try:
        # Parse property
        spec = parseATL(args.property)[0]
        
        # Check property
        satisfied = check(mas, spec)
        
        print('Specification', str(spec), 'is', str(satisfied), file=sys.stderr)
        
        if satisfied:
            print(explain_top(mas, spec).dot())
        else:
            print(cntex_top(mas, spec).dot())
    except ParseException as e:
        print("[ERROR] Cannot parse specification:", str(e))
    except PyNuSMVError as e:
        print("[ERROR]", str(e))
    


"""
CTLK evaluation functions.
"""

from pynusmv.dd import BDD
from pynusmv.mc import eval_simple_expression
from pynusmv.utils import fixpoint as fp

from .ast import (TrueExp, FalseExp, Init, Reachable,
                  Atom, Not, And, Or, Implies, Iff, 
                  AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                  nK, nE, nD, nC, K, E, D, C)


def evalCTLK(fsm, spec):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    
    fsm -- a MAS representing the system
    spec -- an AST-based CTLK specification
    """
    
    if type(spec) is TrueExp:
        return BDD.true(fsm.bddEnc.DDmanager)
        
    elif type(spec) is FalseExp:
        return BDD.false(fsm.bddEnc.DDmanager)
        
    elif type(spec) is Init:
        return fsm.init
        
    elif type(spec) is Reachable:
        return fsm.reachable_states
    
    elif type(spec) is Atom:
        return eval_simple_expression(fsm, spec.value)
        
    elif type(spec) is Not:
        return ~evalCTLK(fsm, spec.child)
        
    elif type(spec) is And:
        return evalCTLK(fsm, spec.left) & evalCTLK(fsm, spec.right)
        
    elif type(spec) is Or:
        return evalCTLK(fsm, spec.left) | evalCTLK(fsm, spec.right)
        
    elif type(spec) is Implies:
        # a -> b = ~a | b
        return (~evalCTLK(fsm, spec.left)) | evalCTLK(fsm, spec.right)
        
    elif type(spec) is Iff:
        # a <-> b = (a & b) | (~a & ~b)
        l = evalCTLK(fsm, spec.left)
        r = evalCTLK(fsm, spec.right)
        return (l & r) | ((~l) & (~r))
        
    elif type(spec) is EX:
        return ex(fsm, evalCTLK(fsm, spec.child))
        
    elif type(spec) is AX:
        # AX p = ~EX ~p
        return ~ex(fsm, ~evalCTLK(fsm, spec.child))
        
    elif type(spec) is EG:
        return eg(fsm, evalCTLK(fsm, spec.child))
        
    elif type(spec) is AG:
        # AG p = ~EF ~p = ~E[ true U ~p ]
        return ~eu(fsm,
                   BDD.true(fsm.bddEnc.DDmanager),
                   ~evalCTLK(fsm, spec.child))
        
    elif type(spec) is EU:
        return eu(fsm, evalCTLK(fsm, spec.left), evalCTLK(fsm, spec.right))
        
    elif type(spec) is AU:
        # A[p U q] = ~E[~q W ~p & ~q] = ~(E[~q U ~p & ~q] | EG ~q)
        p = evalCTLK(fsm, spec.left)
        q = evalCTLK(fsm, spec.right)
        equpq = eu(fsm, ~q, ~q & ~p)
        egq = eg(fsm, ~q)
        return ~(equpq | egq)
        
    elif type(spec) is EF:
        # EF p = E[ true U p ]
        return eu(fsm,
                  BDD.true(fsm.bddEnc.DDmanager),
                  evalCTLK(fsm, spec.child))    
        
    elif type(spec) is AF:
        # AF p = ~EG ~p
        return ~eg(fsm, ~evalCTLK(fsm, spec.child))
        
    elif type(spec) is EW:
        # E[ p W q ] = E[ p U q ] | EG p
        return (eu(fsm, evalCTLK(fsm, spec.left), evalCTLK(fsm, spec.right)) |
                eg(fsm, evalCTLK(fsm, spec.left)))
        
    elif type(spec) is AW:
        # A[p W q] = ~E[~q U ~p & ~q]
        p = evalCTLK(fsm, spec.left)
        q = evalCTLK(fsm, spec.right)
        return ~eu(fsm, ~q, ~p & ~q)
        
    elif type(spec) is nK:
        return nk(fsm, spec.agent.value, evalCTLK(fsm, spec.child))
        
    elif type(spec) is K:
        # K<'a'> p = ~nK<'a'> ~p
        return ~nk(fsm, spec.agent.value, ~evalCTLK(fsm, spec.child))
        
    elif type(spec) is nE:
        return ne(fsm,
                  [a.value for a in spec.group],
                  evalCTLK(fsm, spec.child))
        
    elif type(spec) is E:
        # E<g> p = ~nE<g> ~p
        return ~ne(fsm,
                   [a.value for a in spec.group],
                   ~evalCTLK(fsm, spec.child))
        
    elif type(spec) is nD:
        return nd(fsm,
                  [a.value for a in spec.group],
                  evalCTLK(fsm, spec.child)) 
        
    elif type(spec) is D:
        # D<g> p = ~nD<g> ~p
        return ~nd(fsm,
                   [a.value for a in spec.group],
                   ~evalCTLK(fsm, spec.child))
        
    elif type(spec) is nC:
        return nc(fsm,
                  [a.value for a in spec.group],
                  evalCTLK(fsm, spec.child))
        
    elif type(spec) is C:
        # C<g> p = ~nC<g> ~p
        return ~nc(fsm,
                   [a.value for a in spec.group],
                   ~evalCTLK(fsm, spec.child))
        
    else:
        # TODO Generate error
        print("[ERROR] CTLK evalCTLK: unrecognized specification type", spec)
        return None
    
    
def ex(fsm, phi):
    """
    Return the set of states of fsm satisfying EX phi.
    
    fsm -- a MAS representing the system
    phi -- a BDD representing the set of states of fsm satisfying phi
    """
    return fsm.pre(phi)
    
    
def eg(fsm, phi):
    """
    Return the set of states of fsm satisfying EG phi.
    
    fsm -- a MAS representing the system
    phi -- a BDD representing the set of states of fsm satisfying phi
    """
    return fp(lambda Z: (phi & fsm.pre(Z)),
               BDD.true(fsm.bddEnc.DDmanager))
    
    
def eu(fsm, phi, psi):
    """
    Return the set of states of fsm satisfying E[ phi U psi ].
    
    fsm -- a MAS representing the system
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    """
    return fp(lambda Z: (psi | (phi & fsm.pre(Z))),
               BDD.false(fsm.bddEnc.DDmanager))
    
    
def nk(fsm, agent, phi):
    """
    Return the set of states of fsm satisfying nK<'agent'> phi
    
    fsm -- a MAS representing the system
    agent -- the (str) name of an agent of fsm
    phi -- a BDD representing the set of states of fsm satisfying phi
    """
    # Return the set of states that have a successor in agent's knowledge
    # that satisfies phi and is reachable
    # nK<'a'> 'p' = fsm.equivalent_states(phi, agent)
    return fsm.equivalent_states(phi & fsm.reachable_states,
                                 frozenset({agent})) & fsm.reachable_states
    

def ne(fsm, group, phi):
    """
    Return the set of states of fsm satisfying nE<group> phi
    
    fsm -- a MAS representing the system
    group -- a non-empty list of (str) names of agents of fsm
    phi -- a BDD representing the set of states of fsm satisfying phi
    """
    # A state satisfies nE<g> p iff there is an equivalent state in any relation
    # of any agent in g that satisfies p
    # nE<g> p = \/_{ag in g} nK<ag> p
    result = BDD.false(fsm.bddEnc.DDmanager)
    for agent in group:
        result = result | nk(fsm, agent, phi)
    return result
    
    
def nd(fsm, group, phi):
    """
    Return the set of states of fsm satisfying nD<group> phi
    
    fsm -- a MAS representing the system
    group -- a non-empty list of (str) names of agents of fsm
    phi -- a BDD representing the set of states of fsm satisfying phi
    """
    return (fsm.equivalent_states(phi & fsm.reachable_states, frozenset(group))
            & fsm.reachable_states)
    
    
def nc(fsm, group, phi):
    """
    Return the set of states of fsm satisfying nC<group> phi
    
    fsm -- a MAS representing the system
    group -- a non-empty list of (str) names of agents of fsm
    phi -- a BDD representing the set of states of fsm satisfying phi
    """
    # nC<g> p = mu Z. p | nE<g> Z
    #return fp(lambda Z: (phi | ne(fsm, group, Z)),
    #          BDD.false(fsm.bddEnc.DDmanager))
    # nC<g> p = mu Z. nE<g> (p | Z)
    return fp(lambda Z: ne(fsm, group, (Z | phi)),
              BDD.false(fsm.bddEnc.DDmanager))


from pynusmv_lower_interface.nusmv.parser import parser
from pynusmv.dd import BDD
from pynusmv.mc import eval_ctl_spec

from pynusmv_lower_interface.nusmv.fsm.bdd import bdd as nsBddFsm

def check(fsm, spec, context=None):
    """
    Return whether spec in context is satisfied by fsm.
    """
    violating = (fsm.init & ~eval_ctl(fsm, spec, context=context) &
                 fsm.state_constraints)
    return violating.is_false()

def eval_ctl(fsm, spec, context=None):
    """
    Evaluate spec in fsm.
    
    Return the BDD representing all reachable states of fsm satisfying spec.
    """
    
    if spec.type == parser.CONTEXT:
        sat = eval_ctl(fsm, spec.cdr, spec.car)
        
    elif spec.type == parser.FALSEEXP:
        sat = BDD.false(fsm.bddEnc.DDmanager)
        
    elif spec.type == parser.TRUEEXP:
        sat = BDD.true(fsm.bddEnc.DDmanager)
        
    elif spec.type == parser.NOT:
        sat = ~eval_ctl(fsm, spec.car, context)
        
    elif spec.type == parser.OR:
        left = eval_ctl(fsm, spec.car, context)
        right = eval_ctl(fsm, spec.cdr, context)
        sat = left | right
    
    elif spec.type == parser.AND:
        left = eval_ctl(fsm, spec.car, context)
        right = eval_ctl(fsm, spec.cdr, context)
        sat = left & right
    
    elif spec.type == parser.IMPLIES:
        left = eval_ctl(fsm, spec.car, context)
        right = eval_ctl(fsm, spec.cdr, context)
        sat = ~left | right
        
    elif spec.type == parser.IFF:
        left = eval_ctl(fsm, spec.car, context)
        right = eval_ctl(fsm, spec.cdr, context)
        sat = (left & right) | (~left & ~right)
        
    elif spec.type == parser.EX:
        sat = ex(fsm, eval_ctl(fsm, spec.car, context))
        
    elif spec.type == parser.EF:
        left = BDD.true(fsm.bddEnc.DDmanager)
        right = eval_ctl(fsm, spec.car, context)
        sat = eu(fsm, left, right)
        
    elif spec.type == parser.EG:
        sat = eg(fsm, eval_ctl(fsm, spec.car, context))
        
    elif spec.type == parser.EU:
        sat = eu(fsm,
                 eval_ctl(fsm, spec.car, context),
                 eval_ctl(fsm, spec.cdr, context))
        
    elif spec.type == parser.EW:
        left = eval_ctl(fsm, spec.car, context)
        right = eval_ctl(fsm, spec.cdr, context)
        sat = eg(fsm, left) | eu(fsm, left, right)
        
    elif spec.type == parser.AX:
        left = eval_ctl(fsm, spec.car, context)
        sat = ~ex(fsm, ~left)
        
    elif spec.type == parser.AF:
        left = eval_ctl(fsm, spec.car, context)
        sat = ~eg(fsm, ~left)
        
    elif spec.type == parser.AG:
        left = eval_ctl(fsm, spec.car, context)
        true = BDD.true(fsm.bddEnc.DDmanager)
        sat = ~eu(fsm, true, ~left)
        
    elif spec.type == parser.AU:
        # A[p U q] = !E[!q W (!p & !q)] = !(E[!q U (!p & !q)] | EG !q)
        left = eval_ctl(fsm, spec.car, context)
        right = eval_ctl(fsm, spec.cdr, context)
        sat = ~(eu(fsm, ~right, ~left & ~right) | eg(fsm, ~right))
        
    elif spec.type == parser.AW:
        left = eval_ctl(fsm, spec.car, context)
        right = eval_ctl(fsm, spec.cdr, context)
        sat = ~eu(fsm, ~right, ~left & ~right)
    
    else:
        sat = eval_ctl_spec(fsm, spec)
    
    return sat & fsm.reachable_states

def ex(fsm, phi):
    phi = phi & fsm.reachable_states
    result = fsm.pre(phi)
    return result & fsm.reachable_states

def eg(fsm, phi):
    res = BDD.true(fsm.bddEnc.DDmanager)
    old = BDD.false(fsm.bddEnc.DDmanager)
    while res != old:
        old = res
        new = ex(fsm, res)
        res = res & new & phi & fsm.reachable_states
    return res

def eu(fsm, phi, psi):
    Y = psi & fsm.reachable_states
    old = Y
    new = Y
    while new.isnot_false():
        old = Y
        Y = Y | (ex(fsm, new) & phi)
        new = Y & ~old
    return Y


"""
Extract simple path from a model to explain particular CTLK operators.
"""

from pynusmv.dd import BDD

from .eval import eg, ex, eu, nk, ne, nc, nd

def explain_ex(fsm, state, p):
    """
    Return a path explaining why state of fsm satisfies EX phi.
    
    fsm -- a MAS
    state -- a State of fsm satisfying EX phi
    p -- a BDD representing the set of states of fsm satisfying the property phi
    
    Return a triple (s, i, s') where
        s is state,
        s' is a state of p that is a successor of s,
        i is one possible input between s and s' in fsm
    The returned value represents a two-states path in fsm starting at state
    and ending with a successor of state from p.
    """
    nexts = fsm.post(state)
    sp = fsm.pick_one_state(nexts & p)
    return (state, 
            fsm.pick_one_inputs(fsm.get_inputs_between_states(state, sp)),
            sp)
    

def explain_eg(fsm, state, p):
    """
    Return a path explaining why state of fsm satisfies EG phi.
    
    fsm -- a MAS
    state -- a State of fsm satisfying EG phi
    p -- a BDD representing the set of states of fsm satisfying the property phi
    
    Return a tuple (path, loop) where
        path is a tuple (s_0,..., i_n, s_n) where
            s_0 is state
            s_j belongs to p forall j
            i_j is a possible input between s_j-1 and s_j in fsm
        loop is a tuple (i_l, s_l) where
            s_l belongs to path
            i_l is a possible input between s_n of path and s_l.
    The returned value represents a looping path in fsm composed of states of p
    and starting at state.
    """
    
    allstates = eg(fsm, p)

    # Start path at s
    path = [state]
    # While path[-1] cannot reach itself through states of allstates,
    while (path[-1] & ex(fsm, eu(fsm, allstates, path[-1]))).is_false():
        # choose a successor of path[-1] in allstates and add it in path
        succ = fsm.pick_one_state(fsm.post(path[-1]) & allstates)
        path.append(fsm.pick_one_inputs(
                                fsm.get_inputs_between_states(path[-1], succ)))
        path.append(succ)

    # At this point, path[-1] can reach itself through states of allstates
    # Explain it with explain_eu and explain_ex
    eus = eu(fsm, allstates, path[-1])
    first = explain_ex(fsm, path[-1], eus)
    second = explain_eu(fsm, first[-1], allstates, path[-1])
    fs = tuple(path) + first[1:] + second[1:]

    # Store the loop
    inputs = fs[-2]
    loop = fs[-1]

    # Return the path and the loop
    return (fs[:-2], (inputs, loop))
    

def explain_eu(fsm, state, p, q):
    """
    Return a path explaining why state of fsm satisfies E[ phi U psi ].
    
    fsm -- a MAS
    state -- a State of fsm satisfying E[ phi U psi ]
    p -- a BDD representing the set of states of fsm satisfying the property phi
    q -- a BDD representing the set of states of fsm satisfying the property psi
    
    Return a tuple (s_0, ..., i_n, s_n) where
        s_0 is state
        s_j belongs to p forall j : 0 <= j < n
        s_n belongs to q
        i_j is a possible input between s_j-1 and s_j in fsm.
    The returned value represents a finite path in fsm, ending with a state of q
    with intermediate states of p and starting at state.
    """
    
    # Compute fixpoint and store intermediate BDDs
    funct = lambda Z: (q | (p & fsm.pre(Z)))
    old = BDD.false(fsm.bddEnc.DDmanager)
    new = funct(old)
    paths = [new]
    # Stop when reaching state
    # This is ensured since state satisfies E[ phi U psi ]
    while (state & new).is_false():
        old = new
        new = funct(old)
        paths.append(new - old)
        
    # paths contains intermediate BDDs
    # paths[i] contains the BDD of all states of phi
    # that can reach a state of psi
    # through states of phi, in i steps
    
    # paths[-1] contains state, skip it
    paths = paths[:-1]
    s = state
    path = [s]
    for states in paths[::-1]:
        sp = fsm.pick_one_state(fsm.post(s) & states)
        i = fsm.pick_one_inputs(fsm.get_inputs_between_states(s, sp))
        path.append(i)
        path.append(sp)
        s = sp
    
    return tuple(path)
    

def explain_nk(fsm, state, agent, p):
    """
    Return a couple of states explaining why state satisfies nK<agent> phi.
    
    fsm -- a MAS
    state -- a State of fsm satisfying nK<agent> phi
    agent -- the name of an agent of fsm
    p -- a BDD representing the set of states of fsm satisfying the property phi
    
    Return a tuple (s, ag, s') where
        s is state
        s' belongs to p
        s' is reachable
        s' is equivalent to s for agent in fsm
        ag is {{agent}}.
    The returned value explains why state satisfies nK<agent> phi in fsm
    by showing a state equivalent to state for agent that satisfies phi.
    """
    equiv_states = nk(fsm, agent, state) & fsm.reachable_states
    sp = fsm.pick_one_state(equiv_states & p)
    return (state, frozenset({agent}), sp)
    

def explain_ne(fsm, state, group, p):
    """
    Return a couple of states explaining why state satisfies nE<group> phi.
    
    fsm -- a MAS
    state -- a State of fsm satisfying nE<group> phi
    group -- a list of names of agents of fsm
    p -- a BDD representing the set of states of fsm satisfying the property phi
    
    Return a tuple (s, agent, s') where
        s is state
        s' belongs to p
        s' is equivalent to s for some agent of group in fsm
        agent is the agent of group for which s and s' are equivalent.
    The returned value explains why state satisfies nE<group> phi in fsm
    by showing a state equivalent to state for some agents in group
    that satisfies phi.
    """
    for agent in group:
        nks = nk(fsm, agent, p)
        if state <= nks:
            return explain_nk(fsm, state, agent, p)
    # If this is reached, state |/= nE<group> phi
    

def explain_nd(fsm, state, group, p):
    """
    Return a couple of states explaining why state satisfies nD<group> phi.
    
    fsm -- a MAS
    state -- a State of fsm satisfying nD<group> phi
    group -- a list of names of agents of fsm
    p -- a BDD representing the set of states of fsm satisfying the property phi
    
    Return a tuple (s, group, s') where
        s is state
        s' belongs to p
        s' is distributively equivalent to s for group in fsm.
    The returned value explains why state satisfies nD<group> phi in fsm
    by showing a state group-distributively equivalent to state
    that satisfies phi.
    """
    equiv_states = nd(fsm, group, state) & fsm.reachable_states
    sp = fsm.pick_one_state(equiv_states & p)
    return (state, frozenset(group), sp)
    

def explain_nc(fsm, state, group, p):
    """
    Return a path group's agents equivalent states
    explaining why state satisfies nC<group> phi.
    
    fsm -- a MAS
    state -- a State of fsm satisfying nC<group> phi
    group -- a list of names of agents of fsm
    p -- a BDD representing the set of states of fsm satisfying the property phi
    
    Return a tuple (s_0, ag_1, ..., ag_n, s_n) where
        s_0 is state
        s_n belongs to p
        s_j is equivalent to s_j+1 for some agent in group in fsm,
            forall j : 0 <= j < n
        ag_j+1 is the name of the agent of group
            for which s_j and s_j+1 are equivalent, forall j : 0 <= j < n
    The returned value explains why state satisfies nC<group> phi in fsm
    by showing path in the knowledge of agents of group
    to a state that satisfies phi.
    """
    
    # Compute fixpoint and store intermediate BDDs
    funct = lambda Z: ne(fsm, group, (Z | p))
    old = BDD.false(fsm.bddEnc.DDmanager)
    new = funct(old)
    paths = [new]
    # Stop when reaching state
    # This is ensured since state satisfies nC<group> phi
    while (state & new).is_false():
        old = new
        new = funct(old)
        paths.append(new - old)
    
    # paths[-1] contains state, skip it
    paths = paths[:-1]
    s = state
    path = [s]
    for states in paths[::-1]:
        (olds, ag, s) = explain_ne(fsm, s, group, states)
        path.append(ag)
        path.append(s)
    # Still need to show why the last one satisfies nE<group> phi
    (olds, ag, s) = explain_ne(fsm, path[-1], group, p)
    path.append(ag)
    path.append(s)
    
    return tuple(path)
    
    
def explain_reachable(fsm, state):
    """
    Return a reversed path explaining why state of fsm is reachable.
    
    fsm -- a MAS
    state -- a reachable State of fsm
    
    Return a tuple (s_0, ..., i_n, s_n) where
        s_0 is state
        s_n belongs is an initial state in fsm
        i_j is a possible input between s_j-1 and s_j in fsm
            forall j : 0 < j <= n.
    The returned value represents a finite reversed path in fsm,
    ending with an initial state and starting at state.
    """
    
    # Compute fixpoint and store intermediate BDDs
    funct = lambda Z: (fsm.init | fsm.post(Z))
    old = BDD.false(fsm.bddEnc.DDmanager)
    new = funct(old)
    paths = [new]
    # Stop when reaching state
    # This is ensured since state is reachable
    while (state & new).is_false():
        old = new
        new = funct(old)
        paths.append(new - old)
        
    # paths contains intermediate BDDs
    # paths[i] contains the BDD of all reachable states
    # that can be reached from an initial state in i steps
    
    # paths[-1] contains state, skip it
    paths = paths[:-1]
    s = state
    path = [s]
    for states in paths[::-1]:
        sp = fsm.pick_one_state(fsm.pre(s) & states)
        i = fsm.pick_one_inputs(fsm.get_inputs_between_states(sp, s))
        path.append(i)
        path.append(sp)
        s = sp
    
    return tuple(path)


'''
This module defines a few decorators and (potentially) other useful functions
that makes the writing/reading of test cases clearer and easier to track.
'''
import os
import sys

from pynusmv.init      import init_nusmv, deinit_nusmv
from pynusmv.glob      import load_from_file, master_bool_sexp_fsm
from pynusmv.bmc.glob  import go_bmc, master_be_fsm, bmc_exit

def todo(fn):
    """
    A decorator to tell a test was foreseen but not implemented
    """
    return _log(fn, "Foreseen but not implemented yet")

def skip(reason):
    """
    A decorator to disable a function and warn it has been manually disabled
    """
    return lambda fn: _log(fn, "Manually disabled : {}".format(reason))

def _log(fn, message):
    """
    Returns a function to log the given message
    """
    msg = "INFO : {}.{} -> {}".format(fn.__module__, fn.__name__, message)
    return lambda *x: print(msg, file=sys.stderr) 

def canonical_cnf(be):
    """
    Returns a canonical string representation of the clauses in `be` (enable 
    comparison of generated be's)
    
    :return: a canonical string representation of the clauses in `be` when converted
        in CNF
    """
    cnf = be.to_cnf()
    clit= cnf.formula_literal
    lst = [ sorted([ item for item in array if abs(item) != clit ]) for array in cnf.clauses_list ]
    return str(sorted(lst))

def current_directory(what):
    """
    Returns the current working directory (as of now, cwd) from which the 
    current test case is loaded. Thanks to the cwd, one can write portable test
    cases as the tests will be able to load a resource relative to themselves
    at any time.
    
    Example use::
    
        def setUp(self):
            init_nusmv()
            load_from_file(tests.current_directory(__file__)+"/example.smv")
    
    :param what: the __file__ variable of the ongoing test
    :return: the cwd
    """
    return os.path.dirname(os.path.realpath(what))

class Configure:
    """
    This class provides a convenient context manager to initialize BMC test 
    cases loading the model from a file and setting the most commonly used
    fields in the testcase
    """
    
    def __init__(self, testcase, location, model):
        """
        creates a new instance
        :param testcase: the test case to initialize
        :param location: the __file__ attribute of the module in which the 
            testcase runs
        :param model: a path relative to the testcase module from where to load
            an smv model.
        """
        self.testcase = testcase
        self.model    = current_directory(location)+model
        
    def __enter__(self):
        """Performs what one would usually do in setUp"""
        init_nusmv()
        load_from_file(self.model)
        go_bmc()
        self.testcase.sexpfsm = master_bool_sexp_fsm()
        self.testcase.befsm   = master_be_fsm()
        self.testcase.enc     = self.testcase.befsm.encoding
        self.testcase.mgr     = self.testcase.enc.manager
    
    def __exit__(self, type_, value, traceback):
        """Performs what one would usually do in tearDown"""
        bmc_exit()
        deinit_nusmv()
    


"""
ARCTL explain functions.
"""

from pynusmv.dd import BDD
from .eval import _ex, eag, _eu, eau, eax, evalArctl
from .ast import (TrueExp, FalseExp, Atom, Not, And, Or, Implies, Iff,
                  AaF, AaG, AaX, AaU, AaW, EaF, EaG, EaX, EaU, EaW)
    
    
def explain_witness(fsm, state, spec):
    """
    Explain why state of fsm satisfies spec.
    
    Return a single path explaining (maybe partially) why state of fsm satisfies
    spec. The returned structure is a tuple ((s0, i1, s1, ..., sn), (in, loop))
    where (s0, ..., sn) is a path of fsm explaining spec, and (in, loop)
    represents a loop of this path. If the path is finite, (in, loop)
    is (None, None).
    """
    
    if type(spec) is TrueExp:
        # state is its own explanation
        return ((state,), (None, None))

    elif type(spec) is FalseExp:    
        print("[ERROR] ARCTL explain_witness:",
              "cannot explain why state satisfies False",
              spec)
        # TODO Error, cannot explain why state |= False
        return None
    
    elif type(spec) is Atom:
        # state is its own explanation
        return ((state,), (None, None))
        
    elif type(spec) is Not:
        return explain_countex(fsm, state, spec.child)
        
    elif type(spec) is And:
        # Do not branch. Choose left child
        return explain_witness(fsm, state, spec.left)
        
    elif type(spec) is Or:
        # If state satisfies spec.left, explain it
        # otherwise, state satisfies spec.right, so explain it
        specbdd = evalArctl(fsm, spec.left)
        if state <= specbdd:
            return explain_witness(fsm, state, spec.left)
        else:
            return explain_witness(fsm, state, spec.right)
        
    elif type(spec) is Implies:
        # a -> b is ~a | b
        return explain_witness(fsm, state, Or(Not(spec.left), spec.right))
        
    elif type(spec) is Iff:
        # a <-> b is (a & b) | (~a & ~b)
        return explain_witness(
                fsm, state,
                Or(And(spec.left, spec.right),
                   And(Not(spec.left), Not(spec.right)))
               )
        
    elif type(spec) in {AaF, AaG, AaX, AaU, AaW}:
        # Cannot explain with a single path
        return((state,), (None, None))
                       
    elif type(spec) is EaF:
        path = explain_eau(fsm, state, evalArctl(fsm, spec.action),
                           BDD.true(fsm.bddEnc.DDmanager),
                           evalArctl(fsm, spec.child))
                           
        (npath, loops) = explain_witness(fsm, path[-1], spec.child)
        return (path + npath[1:], loops)
                       
    elif type(spec) is EaG:
        return explain_eag(fsm, state, evalArctl(fsm, spec.action),
                           evalArctl(fsm, spec.child))
                       
    elif type(spec) is EaX:
        path = explain_eax(fsm, state, evalArctl(fsm, spec.action),
                           evalArctl(fsm, spec.child))
        (npath, loops) = explain_witness(fsm, path[-1], spec.child)
        return (path + npath[1:], loops)
                       
    elif type(spec) is EaU:
        path = explain_eau(fsm, state, evalArctl(fsm, spec.action),
                           evalArctl(fsm, spec.left),
                           evalArctl(fsm, spec.right))
                           
        (npath, loops) = explain_witness(fsm, path[-1], spec.right)
        return (path + npath[1:], loops)
                       
    elif type(spec) is EaW:
        # eaw(a, p, q) = ~aau(a, ~q, ~p & ~q)
        return explain_countex(fsm, state,
                                AaU(spec.action, Not(spec.right),
                                    And(Not(spec.left), Not(spec.right))))
        
    else:
        # TODO Generate error
        print("[ERROR] ARCTL explain_witness: unrecognized specification type",
              spec)
        return None
        
        
def explain_countex(fsm, state, spec):
    """
    Explain why state of fsm violates spec.
    
    Return a single path explaining (maybe partially) why state of fsm violates
    spec. The returned structure is a tuple ((s0, i1, s1, ..., sn), (in, loop))
    where (s0, ..., sn) is a path of fsm explaining spec, and (in, loop)
    represents a possible loop of this path. If the path is finite, (in, loop)
    is (None, None).
    """

    if type(spec) is TrueExp:    
        print("[ERROR] ARCTL explain_countex:",
              "cannot explain why state violates True",
              spec)
        # TODO Error, cannot explain why state |/= True
        return None
    
    elif type(spec) is FalseExp:
        # state is its own explanation
        return ((state,), (None, None))
    
    if type(spec) is Atom:
        # state is its own explanation
        return ((state,), (None, None))
        
    elif type(spec) is Not:
        return explain_witness(fsm, state, spec.child)
        
    elif type(spec) is And:
        # ~(a & b) = ~a | ~b
        return explain_witness(fsm, state, Or(Not(spec.left), Not(spec.right)))
        
    elif type(spec) is Or:
        # ~(a | b) = ~a & ~b
        return explain_witness(fsm, state, And(Not(spec.left), Not(spec.right)))
        
    elif type(spec) is Implies:
        # ~(a -> b) = a & ~b
        return explain_witness(fsm, state, And(spec.left, Not(spec.right)))
        
    elif type(spec) is Iff:
        # ~(a <-> b) = (a & ~b) | (~a & b)
        return explain_witness(fsm, state,
                               Or(And(spec.left, Not(spec.right)),
                                  And(Not(spec.left), spec.right)))
        
    elif type(spec) is AaF:
        # ~aaf(a, p) = _eu(a, ~p, ~p & ~_ex(a, true)) | _eg(a, ~p) = eag(a, ~p)
        return explain_witness(fsm, state, EaG(spec.action, Not(spec.child)))
        
    elif type(spec) is AaG:
        return explain_witness(fsm, state, EaF(spec.action, Not(spec.child)))
        
    elif type(spec) is AaX:
        # A<a>X f is false because E<a>X ~f is true or E<a>X true is false
        eaxnf = evalArctl(fsm, EaX(spec.action, Not(spec.child)))
        if state <= eaxnf:
            return explain_witness(fsm, state,
                                   EaX(spec.action, Not(spec.child)))
        else:
            return ((state,), (None, None))
        
    elif type(spec) is AaU:
        return explain_witness(fsm, state,
                               EaW(spec.action,
                                   Not(spec.left),
                                   And(Not(spec.left), Not(spec.right))))
        
    elif type(spec) is AaW:
        return explain_witness(fsm, state,
                               EaU(spec.action,
                                   Not(spec.left),
                                   And(Not(spec.left), Not(spec.right))))
                     
    elif type(spec) in {EaF, EaG, EaX, EaU, EaW}:
        # Cannot explain
        return ((state,), (None, None))
        
    else:
        # TODO Generate error
        print("[ERROR] ARCTL explain_countex: unrecognized specification type",
              spec)
        return None


def explain_eax(fsm, state, alpha, phi):
    """
    Explain why state of fsm satisfies E<a>X p.
    
    fsm -- the BddFsm of the model;
    state -- a State of fsm satisfying E<a>X p;
    alpha -- a BDD of fsm representing the inputs satisfying a;
    phi -- a BDD of fsm representing the states satisfying p.
    
    Return a tuple (state, inputs, state') where
        inputs belongs to alpha
        state' belongs to phi
        state' is reachable from state through inputs.
    """
    
    # Get next state: is a next of state and satisfies phi
    nexts = fsm.post(state, alpha)
    next = fsm.pick_one_state(nexts & phi)
    
    # Get inputs: is an input between state and next and satisfies alpha
    inputs = fsm.get_inputs_between_states(state, next)
    inputs = fsm.pick_one_inputs(inputs & alpha)
    
    return (state, inputs, next)
    
    
def explain_eag(fsm, state, alpha, phi):
    """
    Explain why state of fsm satisfies E<a>G p.
    
    fsm -- the BddFsm of the model;
    state -- a State of fsm satisfying E<a>G p;
    alpha -- a BDD of fsm representing the inputs satisfying a;
    phi -- a BDD of fsm representing the states satisfying p.
    
    Return a tuple (path, (inputs, loop)).
    
    If inputs and loop are None, this means that the witness is a full finite
    path and that path[-1] satisfies ~E<'alpha'>X 'TRUE' (dead state). In this
    case, path is a tuple (s0, i1, ..., sn) where sj |= phi for all j,
    ij |= alpha for all j and sj -ij+1-> sj+1 for all j.
    
    Otherwise, path is a tuple (s0, i1, ..., sn)
    where
        s0 = state
        sj -ij+1-> sj+1 for all j : 0 <= j < n
        sn -inputs-> loop
        loop = sj for some j : 0 <= j <= n
        sj belongs to phi for all j : 0 <= j <= n
        ij belongs to alpha for all j : 0 <= j <= n
        inputs belongs to alpha.
    """
    # eag(a, p) = _eu(a, p, p & ~_ex(a, true)) | _eg(a, p)
    
    paext = phi & ~_ex(fsm, alpha, BDD.true(fsm.bddEnc.DDmanager))
    euppaext = _eu(fsm, alpha, phi, paext)
    
    # If state satisfies _eu(a, p, p & ~_ex(a, true)),
    # use explain_eau to extract such a path
    if state <= euppaext:
        return (explain_eau(fsm, state, alpha, phi, paext), (None, None))
    
    # Otherwise,
    else:    
        # Get allstates = E<alpha>G phi
        allstates = eag(fsm, alpha, phi)
    
        # Start path at s
        path = [state]
        # While path[-1] cannot reach itself through states of eag,
        while (path[-1] &
               eax(fsm,
                   alpha,
                   eau(fsm, alpha, allstates, path[-1]))).is_false():
            # choose a successor of path[-1] in eag and add it in path
            path.append(fsm.pick_one_state(fsm.post(state, alpha) & allstates))
    
        # At this point, path[-1] can reach itself through states of eag
        # Explain it with explain_eau and explain_eax
        eaus = eau(fsm, alpha, allstates, path[-1])
        first = explain_eax(fsm, path[-1], alpha, eaus)
        second = explain_eau(fsm, first[-1], alpha, allstates, path[-1])
        fs = tuple(path) + first[1:] + second[1:]
    
        # Store the loop
        inputs = fs[-2]
        loop = fs[-1]
    
        # Return the path and the loop
        return (fs[:-2], (inputs, loop))
    
    
def explain_eau(fsm, state, alpha, phi, psi):
    """
    Explain why state of fsm satisfies E<a>[p U q].
    
    fsm -- the BddFsm of the model;
    state -- a State of fsm satisfying E<a>[p U q];
    alpha -- a BDD of fsm representing the inputs satisfying a;
    phi -- a BDD of fsm representing the states satisfying p;
    psi -- a BDD of fsm representing the states satisfying q.
    
    Return a tuple (s0, i1, ..., sn) where
        s0 = state
        sj -ij+1-> sj+1 for all j : 0 <= j < n
        sj belongs to phi for all j : 0 <= j < n
        sn belongs to psi
        ij belongs to alpha for all j : 0 <= j <= n.
    """
    
    # Compute fixpoint and store intermediate BDDs
    funct = lambda Z: (psi | (phi & _ex(fsm, alpha, Z)))
    old = BDD.false(fsm.bddEnc.DDmanager)
    new = funct(old)
    paths = [new]
    # Stop when reaching state
    # This is ensured since state satisfies E<a>[p U q]
    while (state & new).is_false():
        old = new
        new = funct(old)
        paths.append(new - old)
        
    # paths contains intermediate BDDs
    # paths[i] contains the BDD of all states of phi
    # that can reach a state of psi
    # through states of phi, in i steps
    # restricted to paths of alpha actions
    
    # paths[-1] contains state, skip it
    paths = paths[:-1]
    s = state
    path = [s]
    for states in paths[::-1]:
        sp = fsm.pick_one_state(fsm.post(s, alpha) & states)
        i = fsm.pick_one_inputs(fsm.get_inputs_between_states(s, sp) & alpha)
        path.append(i)
        path.append(sp)
        s = sp
    
    return tuple(path)


import unittest

from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.fsm import BddFsm
from pynusmv_tools.arctl.check import checkArctl
from pynusmv_tools.arctl.parsing import parseArctl

class TestCheck(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
    
    def tearDown(self):
        deinit_nusmv()
    
    
    def init_model(self):
        fsm = BddFsm.from_filename("tests/pynusmv_tools/arctl/counters.smv")
        self.assertIsNotNone(fsm)
        return fsm
        
    
    def test_atom(self):
        fsm = self.init_model()
        
        spec = parseArctl("'c1.c = 0'")[0]
        self.assertTrue(checkArctl(fsm, spec))
        
        spec = parseArctl("'c2.c = 1'")[0]
        self.assertFalse(checkArctl(fsm, spec)[0])
        
        
    def test_eax(self):
        fsm = self.init_model()
        
        spec = parseArctl("E<'run = rc1'>X 'c1.c = 1'")[0]
        self.assertTrue(checkArctl(fsm, spec)[0])
        
        spec = parseArctl("E<'run = rc2'>X 'c1.c = 1'")[0]
        self.assertFalse(checkArctl(fsm, spec)[0])
        
        spec = parseArctl("E<~'run = rc1'>X 'c1.c = 1'")[0]
        self.assertFalse(checkArctl(fsm, spec)[0])
        
        
    def test_aax(self):
        fsm = self.init_model()
        
        spec = parseArctl("A<'run = rc1'>X 'c1.c = 1'")[0]
        self.assertTrue(checkArctl(fsm, spec)[0])
        
        spec = parseArctl("A<'run = rc1'>X 'c2.c = 1'")[0]
        self.assertFalse(checkArctl(fsm, spec)[0])
        
        
    def test_x(self):
        fsm = self.init_model()
        
        spec = parseArctl("E<'FALSE'>X 'TRUE'")[0]
        self.assertFalse(checkArctl(fsm, spec)[0])
        
        spec = parseArctl("A<'FALSE'>X 'TRUE'")[0]
        self.assertFalse(checkArctl(fsm, spec)[0])
        
        spec = parseArctl("~E<'FALSE'>X 'TRUE'")[0]
        self.assertTrue(checkArctl(fsm, spec)[0])


from .eval import evalArctl
from .explain import explain_witness, explain_countex

def checkArctl(fsm, spec, witness = None, countex = None):
    """
    Check whether fsm satisfies spec.
    
    fsm -- a BddFsm;
    spec -- an ARCTL spec, i.e. an AST made of .ast module classes.
    witness -- None, or a function of signature (fsm, state, spec) that
               returns a witness for fsm, state |= spec.
    countex -- None, or a function of signature (fsm, state, spec) that
               returns a counter-example for fsm, state |= spec
    
    Return a tuple (sat, witness) where
        sat is True if fsm satisfies spec, False otherwise;
        if sat is True,
            witness is the result of witness(fsm, state, spec)
        otherwise,
            witness is the result of countex(fsm, state, spec)
        where state is a state satisfying or violating spec in fsm.
    If witness or countex is None,
        pynusmv_tools.arctl.explain.explain_witness or
        pynusmv_tools.arctl.explain.explain_countex is used.
    """
    init = fsm.init
    specbdd = evalArctl(fsm, spec)
    
    violating = init & ~specbdd
    if violating.isnot_false():
        state = fsm.pick_one_state(violating)
        return (False, countex and countex(fsm, state, spec) or
                       explain_countex(fsm, state, spec))
        
    else:
        satisfying = init & specbdd
        state = fsm.pick_one_state(satisfying)
        return (True, witness and witness(fsm, state, spec) or
                      explain_witness(fsm, state, spec))


"""
AST classes for ARCTL formulas representation.
"""

from collections import namedtuple

# TrueExp, FalseExp
TrueExp = namedtuple('TrueExp', [])
TrueExp.__str__ = lambda self: "True"

FalseExp = namedtuple('FalseExp', [])
FalseExp.__str__ = lambda self: "False"

# Atom
Atom = namedtuple('Atom', ['value'])
Atom.__str__ = lambda self: "'" + str(self.value) + "'"

# logical : Not, And, Or, Implies, Iff
Not = namedtuple('Not', ['child'])
Not.__str__ = lambda self: "~" + "(" + str(self.child) + ")"

And = namedtuple('And', ['left', 'right'])
And.__str__ = lambda self: "(" + str(self.left) + " & " + str(self.right) + ")"

Or = namedtuple('Or', ['left', 'right'])
Or.__str__ = lambda self: "(" + str(self.left) + " | " + str(self.right) + ")"

Implies = namedtuple('Implies', ['left', 'right'])
Implies.__str__ = (lambda self: "(" + str(self.left) + " -> " +
                                      str(self.right) + ")")

Iff = namedtuple('Iff', ['left', 'right'])
Iff.__str__ = (lambda self: "(" + str(self.left) + " <-> " +
                                  str(self.right) + ")")
                                  

# temporal : AaF, AaG, AaX, AaU, AaW, EaF, EaG, EaX, EaU, EaW
AaF = namedtuple('AaF', ['action', 'child'])
AaF.__str__ = (lambda self: "A" + "<" + str(self.action) + ">" +
                            "F " + str(self.child))
                            
AaG = namedtuple('AaG', ['action', 'child'])
AaG.__str__ = (lambda self: "A" + "<" + str(self.action) + ">" +
                            "G " + str(self.child))
                            
AaX = namedtuple('AaX', ['action', 'child'])
AaX.__str__ = (lambda self: "A" + "<" + str(self.action) + ">" +
                            "X " + str(self.child))
                            
AaU = namedtuple('AaU', ['action', 'left', 'right'])
AaU.__str__ = (lambda self: "A" + "<" + str(self.action) + ">" +
                            "[" + str(self.left) + " U " + str(self.right) +
                            "]")
                            
AaW = namedtuple('EaW', ['action', 'left', 'right'])
AaW.__str__ = (lambda self: "A" + "<" + str(self.action) + ">" +
                            "[" + str(self.left) + " W " + str(self.right) +
                            "]")

EaF = namedtuple('EaF', ['action', 'child'])
EaF.__str__ = (lambda self: "E" + "<" + str(self.action) + ">" +
                            "F " + str(self.child))
                            
EaG = namedtuple('EaG', ['action', 'child'])
EaG.__str__ = (lambda self: "E" + "<" + str(self.action) + ">" +
                            "G " + str(self.child))
                            
EaX = namedtuple('EaX', ['action', 'child'])
EaX.__str__ = (lambda self: "E" + "<" + str(self.action) + ">" +
                            "X " + str(self.child))
                            
EaU = namedtuple('EaU', ['action', 'left', 'right'])
EaU.__str__ = (lambda self: "E" + "<" + str(self.action) + ">" +
                            "[" + str(self.left) + " U " + str(self.right) +
                            "]")
                            
EaW = namedtuple('EaW', ['action', 'left', 'right'])
EaW.__str__ = (lambda self: "E" + "<" + str(self.action) + ">" +
                            "[" + str(self.left) + " W " + str(self.right) +
                            "]")


"""
CTLK CLI with TLACE explanation.
"""

import cmd
import argparse
from collections import namedtuple
from pyparsing import ParseException

from pynusmv.init import init_nusmv, deinit_nusmv, reset_nusmv
from pynusmv.exception import PyNuSMVError

from .lazyExplainShell import LazyCTLK_explain_shell
from .explainShell import CTLK_explain_shell

from ..util.nonExitingArgumentParser import (NonExitingArgumentParser,
                                             ArgumentParsingError)
from pynusmv_tools.mas import glob
from ..parsing import parseCTLK
from ..tlace.check import checkCTLK
from ..tlace.xml import xml_witness, xml_countex
from ..lazyTlace.check import checkCTLK as lazyCheckCTLK
from ..simulation.stateChoice import choose_one_state, choose_next_state


Model = namedtuple("Model", ("path", "content"))


class CTLK_shell(cmd.Cmd):
    """
    A shell for CTLK TLACE usage.
    """

    def __init__(self):
        super(CTLK_shell, self).__init__()
        self.prompt = "> "
        self.argparsers = {}
        self.fsm = None
        self.model = None
        self.last = None
        self.paths = []
    
    def preloop(self):
        init_nusmv()
        
    def postloop(self):
        deinit_nusmv()
        
    # For each command cmd
    #   implement parse_cmd : create, store, return the arg parser
    #   implement do_cmd : gets args with , then do the job
    #   implement help_cmd : print arg parser usage
    
    # Commands list
    #   read : read a new model; error if a model is read
    #   fsm : print the path to the read model, or the read model
    #   reset : reset NuSMV, set read model to None
    #   check : check a spec
    #   explain : explain why a spec violated/satisfied
    #   simulate : simulate the model
    #   show_paths : list all the paths already simulated
    # use command arguments to choose aternatives if needed
    
    # Catch PyNuSMVError to show a message on the prompt
    
    def do_model(self, arg):
        if "model" not in self.argparsers:
            self.parse_model()
        
        # Handle arguments parsing errors
        try:
            args = self.argparsers["model"].parse_args(self._split_escaped(arg))
        except ArgumentParsingError as err:
            print(err, end="")
            return False
        
        # Error if
        #   showing path or content when nothing is read
        #   reading a file when one is already read
        
        if (args.content or args.path is None) and self.model is None:
            print("No read model.")
            return False
        
        if args.path is not None and self.model is not None:
            print("model: error: a model is already read from", self.model.path)
            return False
            
        if args.path is None:
            print("read model:", self.model.path)
            
        if args.content:
            print(self.model.content)
            
        if args.path is not None:
            try:
                f = open(args.path, 'r')
                content = f.read()
                f.close()
                glob.load_from_file(args.path)
                self.fsm = glob.mas()
                self.model = Model(args.path, content)
            except PyNuSMVError as err:
                print("model: error:", err)
            except IOError as err:
                print("model: error:", err)
        

    def help_model(self):
        if "model" not in self.argparsers:
            self.parse_model()
        self.argparsers["model"].print_help()
        
    
    def parse_model(self):
        """Build and store the parser for the model command."""
        # model [-c] [PATH]
        # read the model at PATH
        # if PATH is omitted, show the last read model
        # if -c is given, show the content of the read model
        
        if "model" not in self.argparsers:
            parser = NonExitingArgumentParser(
                        "model",
                        description="Read a model.",
                        add_help=False)
            parser.add_argument('path', metavar='PATH',
                                nargs='?', help="the file of the model;"
                                " if omitted, show the current model path")
            parser.add_argument('-c', action='store_true', dest="content",
                                help="show the current model content")
            self.argparsers["model"] = parser
            
            
            
    def do_check(self, arg):
        if "check" not in self.argparsers:
            self.parse_check()
        
        # Error if try to check when no fsm is read
        if self.fsm is None:
            print("check: error: no read model.")
            return False
        
        # Handle arguments parsing error
        try:
            args = self.argparsers["check"].parse_args(self._split_escaped(arg))
        except ArgumentParsingError as err:
            print(err, end="")
            return False
              
        # Check -d and -l options
        if args.diagnostic and args.lazy:
            print("check: error: cannot display a diagnostics if lazy.")
            return False
            
        try:
            spec = parseCTLK(args.spec)[0]
            if not args.lazy:
                (sat, diag) = checkCTLK(self.fsm, spec)
            else:
                (sat, diag) = lazyCheckCTLK(self.fsm, spec)
                
            # Store the result
            self.last = (spec, sat, diag)
            
            print("The formula", '"' + str(spec) + '"', "is {}."
                                            .format("true" if sat else "false"))
            xml = xml_witness if sat else xml_countex
            if args.diagnostic:
                print(xml(self.fsm, diag, spec))
                
        except ParseException as err:
            print("check: error:", err)
        except PyNuSMVError as err:
            print("check: error:", err)
            
        
    def help_check(self):
        if "check" not in self.argparsers:
            self.parse_check()
        self.argparsers["check"].print_help()
        
        # Print CTLK grammar
        print()
        print("""
Grammar for CTLK properties

phi         := atom | logical | temporal | epistemic

logical     := '~'phi | '('phi')' | phi '&' phi | phi '|' phi |
               phi '->' phi | phi '<->' phi
               
temporal    := 'EX' phi | 'EF' phi | 'EG' phi |
               'E[' phi 'U' phi ']' | 'E[' phi 'W' phi ']' |
               'AX' phi | 'AF' phi | 'AG' phi |
               'A[' phi 'U' phi ']' | 'A[' phi 'W' phi ']'
               
epistemic   := 'nK<' agent '>' phi | 'nE<' group '>' phi |
               'nD<' group '>' phi | 'nC<' group '> phi |
               'K<' agent '>' phi | 'E<' group '>' phi |
               'D<' group '>' phi | 'C<' group '> phi

atom being any string surrounded by single quotes (');
agent being an atom and group a comma-separated list of agents.""")
        
        
    def parse_check(self):
        """Build and store the parser for the check command."""
        # check SPEC
        # check the specification SPEC and provide a diagnostic
        
        if "check" not in self.argparsers:
            parser = NonExitingArgumentParser(
                        "check",
                        description="Check a specification.",
                        add_help=False)
            parser.add_argument('spec', metavar='"SPEC"',
                                help="the specification to check")
            parser.add_argument('-d', action='store_true', dest="diagnostic",
                                help="show a diagnostic in XML")
            parser.add_argument('-l', action='store_true', dest="lazy",
                                help="produce a lazy explanation" +
                                     " (cannot be used with -d)")
            self.argparsers["check"] = parser
            
            
    def do_explain(self, arg):
        if "explain" not in self.argparsers:
            self.parse_explain()
        
        # Error if try to check when no fsm is read
        if self.fsm is None:
            print("explain: error: no read model.")
            return False
        
        # Handle arguments parsing error
        try:
            args = (self.argparsers["explain"].
                                           parse_args(self._split_escaped(arg)))
        except ArgumentParsingError as err:
            print(err, end="")
            return False
            
        # Print a warning if lazy and no given specification
        if args.lazy and not args.spec:
            print("explain: warning:",
                  "lazy is ignored since no specification is given.")
            
        try:
            # If spec is not None, check and use the last spec
            # Else check spec
            # Then explain spec by entering the explain cmd
            if args.spec:
                spec = parseCTLK(args.spec)[0]
                if not args.lazy:
                    (sat, diag) = checkCTLK(self.fsm, spec)
                else:
                    (sat, diag) = lazyCheckCTLK(self.fsm, spec)
                self.last = (spec, sat, diag)
                
            if self.last is None:
                print("explain: error: no specification to check.")
                return False
                
            spec = self.last[0]
            sat = self.last[1]
            diag = self.last[2]
            print("The formula", '"' + str(spec) + '"', "is {}."
                                            .format("true" if sat else "false"))
            
            # Start a CTLK explain shell
            if not args.lazy:
                shell = CTLK_explain_shell(diag)
            else:
                shell = LazyCTLK_explain_shell(self.fsm, diag)
            shell.cmdloop()
        except ParseException as err:
            print("check: error:", err)
        except PyNuSMVError as err:
            print("check: error:", err)
            
        
    def help_explain(self):
        if "explain" not in self.argparsers:
            self.parse_explain()
        self.argparsers["explain"].print_help()
        
        
    def parse_explain(self):
        """Build and store the parser for the explain command."""
        # explain ["SPEC"]
        # check the specification SPEC and provide a diagnostic
        # if SPEC is not given, explains the last checked specification
        
        if "explain" not in self.argparsers:
            parser = NonExitingArgumentParser(
                        "explain",
                        description="Check a specification and explain it"
                                    " with an interacive shell.",
                        add_help=False)
            parser.add_argument('spec', metavar='"SPEC"', nargs="?",
                                help="the specification to check;"
                                     " if omitted, explain the last checked "
                                     "specification")
            parser.add_argument('-l', action='store_true', dest="lazy",
                                help="explain with a lazy explanation" +
                                     " (ignored if no given specification)")
            self.argparsers["explain"] = parser
            
            
    def do_simulate(self, arg):
        if "simulate" not in self.argparsers:
            self.parse_simulate()
        
        # Error if try to check when no fsm is read
        if self.fsm is None:
            print("simulate: error: no read model.")
            return False
            
        # Handle arguments parsing error
        try:
            args = (self.argparsers["simulate"].
                                           parse_args(self._split_escaped(arg)))
        except ArgumentParsingError as err:
            print(err, end="")
            return False
            
        # Check existence of a under stateid, if given
        if args.stateid is not None:
            # args.stateid should be path.state
            # Check it and put it in state variable
            sp = args.stateid.split(".")
            if len(sp) != 2:
                print("simulate: error: ID should be a couple pathId.stateId.")
                return False
            try:
                pathid = int(sp[0])
                stateid = int(sp[1])
                if pathid <= 0 or len(self.paths) < pathid:
                    print("simulate: error: {} is not a valid path ID."
                                                                .format(pathid))
                    return False
                if (stateid <= 0
                    or len(self.paths[pathid - 1]) <= (stateid - 1) * 2):
                    print("simulate: error: {}"
                          " is not a valid state ID in path {}."
                          .format(stateid, pathid))
                    return False
                state = self.paths[pathid - 1][(stateid - 1) * 2]
                self._show_path([state])
            except ValueError:    
                # Print an error and return if the state is misspecified
                print("simulate: error: ID should be a couple path.state.")
                return False
        else:
            state = choose_one_state(self.fsm, self.fsm.init)
            if state is None:
                # No chosen state, abort
                return False
        
        path = [state]
        while state is not None:
            print("----- Choose next state")
            (inputs, state) = choose_next_state(self.fsm, state)
            if state is None:
                self.paths.append(path)
                return False
            else:
                path.append(inputs)
                path.append(state)
        
        
    def help_simulate(self):
        if "simulate" not in self.argparsers:
            self.parse_simulate()
        self.argparsers["simulate"].print_help()
        
        
    def parse_simulate(self):
        """Build and store the parser of the simulate command."""
        # simulate [path.state]
        # simulate the model from path.state,
        # i.e. the stateth state of pathth path, if path.state is given,
        # simulate the model and start by asking for an initial state otherwise
        
        if "simulate" not in self.argparsers:
            parser = NonExitingArgumentParser(
                        "simulate",
                        description="Simulate the model.",
                        add_help=False)
            parser.add_argument('stateid', metavar='"ID"', nargs="?",
                                help="the id of the starting state, "
                                     "as pathId.stateId; "
                                     "if omitted, "
                                     "start by asking for an initial sate")
            self.argparsers["simulate"] = parser
        
        
    def do_show_path(self, arg):
        if "show_path" not in self.argparsers:
            self.parse_show_path()
        
        # Error if try to check when no fsm is read
        if self.fsm is None:
            print("show_path: error: no read model.")
            return False
            
        # Handle arguments parsing error
        try:
            args = (self.argparsers["show_path"].
                                           parse_args(self._split_escaped(arg)))
        except ArgumentParsingError as err:
            print(err, end="")
            return False
            
        if args.pathid is not None:
            if args.pathid <= 0 or len(self.paths) < args.pathid:
                print("show_path: error: {} is not a valid path ID."
                                                           .format(args.pathid))
                return False
            
            else:
                # Show path
                path = self.paths[args.pathid - 1]
                self._show_path(path, args.pathid)
                return False
        else:
            # Show all
            i = 1
            for path in self.paths:
                header = (" Path " + str(i) +
                          " (" + str(int(len(path) / 2 + 1)) + " states) ")
                print(header.center(40, "-"))
                self._show_path(path[0:1])
                i += 1
            return False            
        
        
    def help_show_path(self):
        if "show_path" not in self.argparsers:
            self.parse_show_path()
        self.argparsers["show_path"].print_help()
        
        
    def parse_show_path(self):
        """Build and store the parser of the show_path command."""
        # show_path [path]
        # show the given path,
        # show all the paths (first state + length) if path is omitted
        
        if "show_path" not in self.argparsers:
            parser = NonExitingArgumentParser(
                        "show_path",
                        description="Show simulated paths.",
                        add_help=False)
            parser.add_argument('pathid', metavar='"PATHID"', nargs="?",
                                type=int,
                                help="the id of the path to show; "
                                     "if omitted, "
                                     "show all the paths")
            self.argparsers["show_path"] = parser
            
            
    def do_reset(self, args):
        """Reset the prompt; forget the read model and restart NuSMV."""
        self.fsm = None
        self.model = None
        self.last = None
        self.paths = []
        glob.reset_globals()
        reset_nusmv()
    
    
    def do_EOF(self, arg):
        """Quit the prompt (you can hit CTRL-d)."""
        return True
            
            
    def do_quit(self, arg):
        """Quit the prompt."""
        return True
        
        
    def _split_escaped(self, string, sep=None, escape=None):
        """
        Split string with each occurrence of sep,
        except if this occurrence is between two occurrences of escape.
        """
        sep = sep if sep else [" "]
        escape = escape if escape else '"'
        
        res = []
        cur = ""
        esc = False
        for char in string:
            if char == escape:
                esc = not esc
            elif char in sep and not esc:
                if cur != "":
                    res.append(cur)
                    cur = ""
            else:
                cur += char
        if cur != "":
            res.append(cur)
        return res
        
        
    def _show_path(self, path, pathprefix=None):
        """
        Show the given path.
        
        path -- a list of states separated by inputs.
        """
        def show_elem(elem, prev=None):
            values = elem.get_str_values()
            for var in values:
                if (prev is not None and
                    prev.get_str_values()[var] != values[var] or
                    prev is None) :
                    print(var, "=", values[var])
        
        i = 1
        if len(path) > 0:
            header = (" State " +
                      ((str(pathprefix) + "." + str(i)) + " " if pathprefix
                        else ""))
            print(header.center(40, "-"))
            show_elem(path[0])
            prev = path[0]
        else:
            prev = None
        for (inputs, state) in zip(path[1::2], path[2::2]):
            # Show inputs
            if inputs is not None:
                header = " Inputs "
                print(header.center(40, "-"))
                show_elem(inputs)
            
            # Show state
            i += 1
            header = (" State " +
                      ((str(pathprefix) + "." + str(i)) if pathprefix else "") +
                      " ")
            print(header.center(40, "-"))
            show_elem(state, prev)
                    
            prev = state

def main():
    CTLK_shell().cmdloop()

if __name__ == "__main__":
    main()



'''
This test module validates the behavior of the SAT problem generation 
translating the bounded semantics verification of LTL properties as defined 
in :mod:`pynusmv_tools.bmcLTL.gen`
'''
from unittest                    import TestCase
from tests                       import utils as tests
from pynusmv_tools.bmcLTL                import gen

from pynusmv.bmc                 import ltlspec
from pynusmv.bmc                 import utils as bmcutils
from pynusmv.parser              import parse_ltl_spec
from pynusmv.node                import Node
from pynusmv.be.expression       import Be 
from pynusmv_tools.bmcLTL.parsing        import parseLTL
from pynusmv.sat                 import SatSolverFactory, Polarity,\
    SatSolverResult

class TestGen(TestCase):
 
    def test_model_problem(self):
        with tests.Configure(self, __file__, "/example.smv"):
            enc = self.enc
            # no step taken
            bound = 0
             
            tool  = gen.model_problem(self.befsm, bound)
            manual= enc.shift_to_time(self.befsm.init, 0)
             
            nusmv = bmcutils.BmcModel().path(bound, with_init=True)
             
            s_tool   = tests.canonical_cnf(tool)
            s_manual = tests.canonical_cnf(manual)
            s_nusmv  = tests.canonical_cnf(nusmv)
             
            self.assertEqual(s_tool, s_nusmv)
            self.assertEqual(s_tool, s_manual)
             
            # one step taken
            bound = 1
             
            tool  = gen.model_problem(self.befsm, bound)
            manual= enc.shift_to_time(self.befsm.trans,0) &\
                    enc.shift_to_time(self.befsm.init, 0)
             
            nusmv = bmcutils.BmcModel().path(bound, with_init=True)
             
            s_tool   = tests.canonical_cnf(tool)
            s_manual = tests.canonical_cnf(manual)
            s_nusmv  = tests.canonical_cnf(nusmv)
             
            self.assertEqual(s_tool, s_manual)
            self.assertEqual(s_tool, s_nusmv)
             
            # two steps
            bound = 2
             
            tool  = gen.model_problem(self.befsm, bound)
            manual= enc.shift_to_time(self.befsm.init, 0) &\
                    enc.shift_to_time(self.befsm.trans,0) &\
                    enc.shift_to_time(self.befsm.trans,1)
             
            nusmv = bmcutils.BmcModel().path(bound, with_init=True)
             
            s_tool   = tests.canonical_cnf(tool)
            s_manual = tests.canonical_cnf(manual)
            s_nusmv  = tests.canonical_cnf(nusmv)
             
            self.assertEqual(s_tool, s_manual)
            self.assertEqual(s_tool, s_nusmv)
    
    def verify_invariants_constraint(self, bound):
        model  = bmcutils.BmcModel() 
            
        manual = Be.true(self.mgr) 
        for i in range(bound+1):
            manual &= model.invar[i]
        
        tool = gen.invariants_constraint(self.befsm, bound)
        
        self.assertEqual(tests.canonical_cnf(tool), 
                         tests.canonical_cnf(manual))
        
    def test_invariants_constraint(self):
        with tests.Configure(self, __file__, "/example.smv"):
            self.verify_invariants_constraint(0)
            self.verify_invariants_constraint(1)
            self.verify_invariants_constraint(2)
            self.verify_invariants_constraint(3)
        
        with tests.Configure(self, __file__, "/dummy_with_invar.smv"):
            self.verify_invariants_constraint(0)
            self.verify_invariants_constraint(1)
            self.verify_invariants_constraint(2)
            self.verify_invariants_constraint(3)
    
    def satisfiability(self, problem):
        solver = SatSolverFactory.create()
        solver+= problem.to_cnf()
        solver.polarity(problem.to_cnf(), Polarity.POSITIVE)
        return solver.solve()
       
    def validate_generate_problem(self, bound, custom_text, nusmv_text):
        fsm     = self.befsm
        # formulae
        formula = parseLTL(custom_text)
        fml_node= Node.from_ptr(parse_ltl_spec(nusmv_text))
         
        # IMPORTANT NOTE: each instantiation of the problem creates new CNF 
        #   literal which appears in the clauses list (even in canonical form)
        #   hence, the canonical forms of the different instantiations cannot
        #   simply be compared as there is no a priori way to know what CNF 
        #   literal reconcile with what other.
        #   However, the different expressions should all have the exact same
        #   satisfiability. So, that's how this test proceeds.
          
        smv     = ltlspec.generate_ltl_problem(fsm, fml_node, bound)
        tool    = gen.generate_problem(formula, fsm, bound)
        manual  = gen.model_problem(fsm, bound) &\
                  formula.nnf(True).bounded_semantics(fsm, bound)
         
        sat_smv = self.satisfiability(smv)
        sat_tool= self.satisfiability(tool)
        sat_man = self.satisfiability(manual)
         
        self.assertEqual(sat_tool, sat_man)
        self.assertEqual(sat_tool, sat_smv)
         
    def test_generate_problem_nofairness_noinvars(self):
        with tests.Configure(self, __file__, "/example.smv"):
            # length 0
            self.validate_generate_problem(0, "<>(a <=> !b)", "F (a <-> !b)")
            self.validate_generate_problem(0, "[](a <=> !b)", "G (a <-> !b)")
            self.validate_generate_problem(0, "(a U b)", "(a U b)")
            self.validate_generate_problem(0, "a => () b", "a -> (X b)")
            # length 1
            self.validate_generate_problem(1, "<>(a <=> !b)", "F (a <-> !b)")
            self.validate_generate_problem(1, "[](a <=> !b)", "G (a <-> !b)")
            self.validate_generate_problem(1, "(a U b)", "(a U b)")
            self.validate_generate_problem(1, "a => () b", "a -> (X b)")
            # length 2
            self.validate_generate_problem(2, "<>(a <=> !b)", "F (a <-> !b)")
            self.validate_generate_problem(2, "[](a <=> !b)", "G (a <-> !b)")
            self.validate_generate_problem(2, "(a U b)", "(a U b)")
            self.validate_generate_problem(2, "a => () b", "a -> (X b)")
            # length 3
            self.validate_generate_problem(3, "<>(a <=> !b)", "F (a <-> !b)")
            self.validate_generate_problem(3, "[](a <=> !b)", "G (a <-> !b)")
            self.validate_generate_problem(3, "(a U b)", "(a U b)")
            self.validate_generate_problem(3, "a => () b", "a -> (X b)")
        
    def test_generate_problem_with_fairness(self):
        '''
        This test clearly shows the difference in validating a property with
        or without fairness constraint
        '''
        with tests.Configure(self, __file__, "/philo.smv"):
            # length 0
            # nusmv has fairness always on.
            fml_node= Node.from_ptr(parse_ltl_spec("G (p1.waiting -> F !p1.waiting)"))
            smv     = ltlspec.generate_ltl_problem(self.befsm, fml_node, 0)
            self.assertEqual(SatSolverResult.UNSATISFIABLE, self.satisfiability(smv))
            
            formula = parseLTL("[](p1.waiting => <>!p1.waiting)")
            unfair  = gen.generate_problem(formula, self.befsm, 0, no_fairness=True)
            self.assertEqual(SatSolverResult.UNSATISFIABLE, self.satisfiability(unfair))
            
            fair    = gen.generate_problem(formula, self.befsm, 0, no_fairness=False)
            self.assertEqual(SatSolverResult.UNSATISFIABLE, self.satisfiability(fair))
            
            # length 1
            fml_node= Node.from_ptr(parse_ltl_spec("G (p1.waiting -> F !p1.waiting)"))
            smv     = ltlspec.generate_ltl_problem(self.befsm, fml_node, 1)
            self.assertEqual(SatSolverResult.UNSATISFIABLE, self.satisfiability(smv))
            
            formula = parseLTL("[](p1.waiting => <>!p1.waiting)")
            unfair  = gen.generate_problem(formula, self.befsm, 1, no_fairness=True)
            self.assertEqual(SatSolverResult.SATISFIABLE, self.satisfiability(unfair))
            
            fair    = gen.generate_problem(formula, self.befsm, 1, no_fairness=False)
            self.assertEqual(SatSolverResult.UNSATISFIABLE, self.satisfiability(fair))
            
    def test_generate_problem_with_invars(self):
        '''
        This test clearly shows the difference in validating a property with
        or without fairness constraint
        '''
        with tests.Configure(self, __file__, "/dummy_with_invar.smv"):
            # length 0
            formula = parseLTL("[] v")
            
            noinvar = gen.generate_problem(formula, self.befsm, 0, no_invar=True)
            self.assertEqual(SatSolverResult.SATISFIABLE, self.satisfiability(noinvar))
            
            w_invar = gen.generate_problem(formula, self.befsm, 0, no_invar=False)
            self.assertEqual(SatSolverResult.UNSATISFIABLE, self.satisfiability(w_invar))
            
            # length 1
            noinvar = gen.generate_problem(formula, self.befsm, 1, no_invar=True)
            self.assertEqual(SatSolverResult.SATISFIABLE, self.satisfiability(noinvar))
            
            w_invar = gen.generate_problem(formula, self.befsm, 1, no_invar=False)
            self.assertEqual(SatSolverResult.UNSATISFIABLE, self.satisfiability(w_invar))
            


"""
ATLK with partial observability evaluation functions.

These evaluation functions reimplement the algorithms and try to be more
optimal.
"""

import gc

from pynusmv.dd import BDD
from pynusmv.mc import eval_simple_expression
from pynusmv.utils import fixpoint as fp
from pynusmv.exception import PyNuSMVError

from ..atlkFO.ast import (TrueExp, FalseExp, Init, Reachable,
                          Atom, Not, And, Or, Implies, Iff, 
                          AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                          nK, nE, nD, nC, K, E, D, C,
                          CEF, CEG, CEX, CEU, CEW, CAF, CAG, CAX, CAU, CAW)
                          

from ..atlkFO.eval import (fair_states, ex, eg, eu, nk, ne, nd, nc)


def evalATLK(fsm, spec, variant="SF", semantics="group"):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    
    fsm -- a MAS representing the system
    spec -- an AST-based ATLK specification
    variant -- the variant of the algorithm to evaluate strategic operators;
               must be
               * "SF" for the standard way: splitting in uniform strategies then
                 filtering winning states,
               * "FS" for the alternating way: filtering winning states, then
                 splitting one conflicting equivalence class, then recurse
               * "FSF" for the filter-split-filter way: filtering winning states
                 then splitting all remaining actions into uniform strategies,
                 then filtering final winning states.
                 
    If variant is not in {"SF", "FS", "FSF"}, the standard "SF" way is used.
    """
    
    if semantics != "group":
        raise PyNuSMVError("Optimal evalATLK: unsupported semantics:" +
                           semantics)
    
    if type(spec) is TrueExp:
        return BDD.true(fsm.bddEnc.DDmanager)
        
    elif type(spec) is FalseExp:
        return BDD.false(fsm.bddEnc.DDmanager)
        
    elif type(spec) is Init:
        return fsm.init
        
    elif type(spec) is Reachable:
        return fsm.reachable_states
    
    elif type(spec) is Atom:
        return eval_simple_expression(fsm, spec.value)
        
    elif type(spec) is Not:
        return ~evalATLK(fsm, spec.child, variant=variant)
        
    elif type(spec) is And:
        return (evalATLK(fsm, spec.left, variant=variant)
                & evalATLK(fsm, spec.right, variant=variant))
        
    elif type(spec) is Or:
        return (evalATLK(fsm, spec.left, variant=variant)
                | evalATLK(fsm, spec.right, variant=variant))
        
    elif type(spec) is Implies:
        # a -> b = ~a | b
        return ((~evalATLK(fsm, spec.left, variant=variant))
                 | evalATLK(fsm, spec.right, variant=variant))
        
    elif type(spec) is Iff:
        # a <-> b = (a & b) | (~a & ~b)
        l = evalATLK(fsm, spec.left, variant=variant)
        r = evalATLK(fsm, spec.right, variant=variant)
        return (l & r) | ((~l) & (~r))
        
    elif type(spec) is EX:
        return ex(fsm, evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is AX:
        # AX p = ~EX ~p
        return ~ex(fsm, ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is EG:
        return eg(fsm, evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is AG:
        # AG p = ~EF ~p = ~E[ true U ~p ]
        return ~eu(fsm,
                   BDD.true(fsm.bddEnc.DDmanager),
                   ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is EU:
        return eu(fsm, evalATLK(fsm, spec.left, variant=variant),
                       evalATLK(fsm, spec.right, variant=variant))
        
    elif type(spec) is AU:
        # A[p U q] = ~E[~q W ~p & ~q] = ~(E[~q U ~p & ~q] | EG ~q)
        p = evalATLK(fsm, spec.left, variant=variant)
        q = evalATLK(fsm, spec.right, variant=variant)
        equpq = eu(fsm, ~q, ~q & ~p)
        egq = eg(fsm, ~q)
        return ~(equpq | egq)
        
    elif type(spec) is EF:
        # EF p = E[ true U p ]
        return eu(fsm,
                  BDD.true(fsm.bddEnc.DDmanager),
                  evalATLK(fsm, spec.child, variant=variant))    
        
    elif type(spec) is AF:
        # AF p = ~EG ~p
        return ~eg(fsm, ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is EW:
        # E[ p W q ] = E[ p U q ] | EG p
        return (eu(fsm, evalATLK(fsm, spec.left, variant=variant),
                        evalATLK(fsm, spec.right, variant=variant)) |
                eg(fsm, evalATLK(fsm, spec.left, variant=variant)))
        
    elif type(spec) is AW:
        # A[p W q] = ~E[~q U ~p & ~q]
        p = evalATLK(fsm, spec.left, variant=variant)
        q = evalATLK(fsm, spec.right, variant=variant)
        return ~eu(fsm, ~q, ~p & ~q)
        
    elif type(spec) is nK:
        return nk(fsm, spec.agent.value, evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is K:
        # K<'a'> p = ~nK<'a'> ~p
        return ~nk(fsm, spec.agent.value, ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is nE:
        return ne(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is E:
        # E<g> p = ~nE<g> ~p
        return ~ne(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is nD:
        return nd(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant)) 
        
    elif type(spec) is D:
        # D<g> p = ~nD<g> ~p
        return ~nd(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is nC:
        return nc(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is C:
        # C<g> p = ~nC<g> ~p
        return ~nc(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant))
                   
    elif type(spec) in {CEX, CAX, CEG, CAG, CEU, CAU, CEF, CAF, CEW, CAW}:
        if variant == "SF":
            return eval_strat(fsm, spec)
        elif variant == "FS":
            return eval_strat_improved(fsm, spec)
        elif variant == "FSF":
            return eval_strat_FSF(fsm, spec)
        else:
            return eval_strat(fsm, spec)
        
    else:
        # TODO Generate error
        print("[ERROR] evalATLK: unrecognized specification type", spec)
        return None
        
        
def split_one(fsm, strats, gamma):
    """
    Split one equivalence class of strats and return triples composed of
    the common non-conflicting part already encountered,
    a split and the rest of strats to split.
    
    fsm -- the model
    strats -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    
    Return a generator of all triples of common parts, splits
    and rest of strats.
    
    """
    if strats.is_false():
        yield (strats, strats, strats)
        return
        
    else:
        ngamma_cube = fsm.bddEnc.inputsCube - fsm.inputs_cube_for_agents(gamma)
        common = BDD.false(fsm.bddEnc.DDmanager)
        
        while strats.isnot_false():
            # Get one equivalence class
            si = fsm.pick_one_state_inputs(strats)
            s = si.forsome(fsm.bddEnc.inputsCube)
            eqs = fsm.equivalent_states(s, gamma)
            eqcl = strats & eqs
            
            # Remove it from strats
            strats = strats - eqcl
            
            # The current equivalence class is conflicting
            if((eqcl - (eqcl & si.forsome(fsm.bddEnc.statesCube | ngamma_cube)))
                .isnot_false()):
                # Split eqcl into non-conflicting subsets
                while eqcl.isnot_false():
                    si = fsm.pick_one_state_inputs(eqcl)
                    ncss = eqcl & si.forsome(fsm.bddEnc.statesCube |ngamma_cube)
                    eqcl = eqcl - ncss

                    yield (common, ncss, strats)
                return
            
            else:
                # Add equivalence class to common
                common = common | eqcl
        
        # strats is false, everything is in common
        yield (common, strats, strats)


def split_eval(fsm, spec, common, rest):
    """
    Split strats into all its non-conflicting greatest subsets and compute
    the states that satisfy spec in strats.
    
    fsm -- the model
    strats -- a BDD representing a set of states/inputs pairs
    spec -- an AST-based ATLK specification with a top strategic operator.
    
    """    
    if (common | rest).is_false():
        return common | rest
    else:
        gamma = {atom.value for atom in spec.group}
        sat = BDD.false(fsm.bddEnc.DDmanager)
        
        for newcommon, splitted, newrest in split_one(fsm, rest, gamma):
            if splitted.is_false():
                # newcommon contains a partial uniform strategy
                # we can extract the winning states
                # from newcommon | common
                
                # first, filter
                winning = (filter_strat(fsm, spec, common | newcommon, "SF")
                           .forsome(fsm.bddEnc.inputsCube))
                
                # then, get equiv sat
                sat = sat | all_equiv_sat(fsm, winning, gamma)
                gc.collect()
            else:
                sat = sat | split_eval(fsm, spec, common | newcommon | splitted,
                                       newrest)
        return sat


def all_equiv_sat(fsm, winning, agents):
    """
    Return the states s of winning such that all states indistinguishable
    from s by agents are in winning.
    
    fsm -- a MAS representing the system;
    winning -- a BDD representing a set of states of fsm;
    agents -- a set of agents.
    
    """
    # Get states for which all states belong to winning
    # wineq is the set of states for which all equiv states are in winning
    nwinning = ~winning & fsm.bddEnc.statesInputsMask
    return ~(fsm.equivalent_states(nwinning &
             fsm.reachable_states, frozenset(agents))) & winning


def eval_strat(fsm, spec):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    spec is a strategic operator <G> pi.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator.
    
    """
    if type(spec) is CAX:
        # [g] X p = ~<g> X ~p
        newspec = CEX(spec.group, Not(spec.child))
        return ~eval_strat(fsm, newspec)
        
    elif type(spec) is CAG:
        # [g] G p = ~<g> F ~p
        newspec = CEF(spec.group, Not(spec.child))
        return ~eval_strat(fsm, newspec)
        
    elif type(spec) is CAU:
        # [g][p U q] = ~<g>[ ~q W ~p & ~q ]
        newspec = CEW(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~eval_strat(fsm, newspec)
        
    elif type(spec) is CAF:
        # [g] F p = ~<g> G ~p
        newspec = CEG(spec.group, Not(spec.child))
        return ~eval_strat(fsm, newspec)
        
    elif type(spec) is CAW:
        # [g][p W q] = ~<g>[~q U ~p & ~q]
        newspec = CEU(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~eval_strat(fsm, newspec)
    
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    
    return split_eval(fsm, spec, BDD.false(fsm.bddEnc.DDmanager),
                      fsm.protocol(agents) & fsm.reachable_states)


def cex_si(fsm, agents, phi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents> X phi
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    strat -- a BDD representing allowed state/inputs pairs, or None
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask
    
    return fsm.pre_strat_si(phi | nfair_gamma_si(fsm, agents, strat),
                            agents, strat)
    

def ceu_si(fsm, agents, phi, psi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents>[phi U psi]
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    psi = psi & fsm.bddEnc.statesInputsMask & strat
    
    if len(fsm.fairness_constraints) == 0:
        return fp(lambda Z : psi | (phi & fsm.pre_strat_si(Z, agents, strat)),
                  BDD.false(fsm.bddEnc.DDmanager))
    else:
        nfair = nfair_gamma_si(fsm, agents, strat)
        def inner(Z):
            res = psi
            for f in fsm.fairness_constraints:
                nf = ~f & fsm.bddEnc.statesMask & strat
                res = res | fsm.pre_strat_si(fp(lambda Y :
                                                 (phi | psi | nfair) &
                                                 (Z | nf) &
                                                 (psi |
                                                  fsm.pre_strat_si(Y, agents,
                                                                   strat)
                                                 ),
                                              BDD.true(fsm.bddEnc.DDmanager)),
                                              agents, strat)
            return (psi | phi | nfair) & res
        return fp(inner, BDD.false(fsm.bddEnc.DDmanager))
    

def cew_si(fsm, agents, phi, psi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents>[phi W psi]
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    psi = psi & fsm.bddEnc.statesInputsMask & strat
    
    nfair = nfair_gamma_si(fsm, agents, strat)
    
    return fp(lambda Y : (psi | phi | nfair) &
                         (psi | fsm.pre_strat_si(Y, agents, strat)),
              BDD.true(fsm.bddEnc.DDmanager))
    
    
def ceg_si(fsm, agents, phi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents> G phi
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    
    nfair = nfair_gamma_si(fsm, agents, strat)
    
    return fp(lambda Y : (phi | nfair) & fsm.pre_strat_si(Y, agents, strat),
              BDD.true(fsm.bddEnc.DDmanager))


def nfair_gamma_si(fsm, agents, strat=None):
    """
    Return the set of state/inputs pairs of strat
    in which agents can avoid a fair path in strat.
    If strat is None, it is considered true.
    
    fsm -- the model
    agents -- a list of agents names
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    if len(fsm.fairness_constraints) == 0:
        return BDD.false(fsm.bddEnc.DDmanager)
    else:
        def inner(Z):
            res = BDD.false(fsm.bddEnc.DDmanager)
            for f in fsm.fairness_constraints:
                nf = ~f & fsm.bddEnc.statesMask & strat
                res = res | fsm.pre_strat_si(fp(lambda Y :
                                                 (Z | nf) &
                                                 fsm.pre_strat_si(Y, agents,
                                                                  strat),
                                             BDD.true(fsm.bddEnc.DDmanager)),
                                             agents, strat)
            return res
        return fp(inner, BDD.false(fsm.bddEnc.DDmanager))



def eval_strat_improved(fsm, spec, toSplit=None, toKeep=None):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    spec is a strategic operator <G> pi.
    
    Implement the improved algorithm alternating between splitting a conflicting
    equvalence class and filtering winning state/inputs pairs.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator.
    toSplit -- the part of the strategy to split (not necessarily uniform).
    toKeep -- the part of the strategy to keep (uniform).
    
    """
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    gamma = agents
    ngamma_cube = fsm.bddEnc.inputsCube - fsm.inputs_cube_for_agents(gamma)
    
    if not toSplit and not toKeep:
        toSplit = fsm.protocol(agents)
        toKeep = BDD.false(fsm.bddEnc.DDmanager)
    
    if type(spec) is CAX:
        # [g] X p = ~<g> X ~p
        newspec = CEX(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, "FS")

    elif type(spec) is CAG:
        # [g] G p = ~<g> F ~p
        newspec = CEF(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, "FS")

    elif type(spec) is CAU:
        # [g][p U q] = ~<g>[ ~q W ~p & ~q ]
        newspec = CEW(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~evalATLK(fsm, newspec, "FS")

    elif type(spec) is CAF:
        # [g] F p = ~<g> G ~p
        newspec = CEG(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, "FS")

    elif type(spec) is CAW:
        # [g][p W q] = ~<g>[~q U ~p & ~q]
        newspec = CEU(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~evalATLK(fsm, newspec, "FS")
    
    winning = filter_strat(fsm, spec, toSplit | toKeep, variant="FS")
    toSplit = toSplit & winning
    toKeep = toKeep & winning
    
    # Get one conflicting equivalence class
    if winning.is_false(): # no state/inputs pairs are winning => return false
        return winning
    
    # TODO Improve this: if a pair is not conflicting now, it won't be 
    # conflicting anymore. We should keep track of non-conflicting equivalence
    # classes through recursive calls
    
    # Split one conflicting equivalence class
    for common, splitted, rest in split_one(fsm, toSplit, gamma):
        
        if splitted.is_false():
            # No conflicting classes, return states that are winning for all eq
            common = (common | toKeep).forsome(fsm.bddEnc.inputsCube)
            sat = sat | all_equiv_sat(fsm, common, agents)
        
        else:
            sat = sat | eval_strat_improved(fsm, spec,
                                            toSplit=rest,
                                            toKeep=toKeep | common | splitted)
    
    return sat


def filter_strat(fsm, spec, strat=None, variant="SF"):
    """
    Returns the subset SA of strat (or the whole system if strat is None),
    state/action pairs of fsm, such that there is a strategy to satisfy spec
    in fsm.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator;
            the operator is CEX, CEG, CEF, CEU or CEW.
    strat -- the subset of the system to consider.
    variant -- the variant of the algorithm to evaluate strategic operators;
               must be
               * "SF" for the standard way: splitting in uniform strategies then
                 filtering winning states,
               * "FS" for the alternating way: filtering winning states, then
                 splitting one conflicting equivalence class, then recurse
               * "FSF" for the filter-split-filter way: filtering winning states
                 then splitting all remaining actions into uniform strategies,
                 then filtering final winning states.
                 
    If variant is not in {"SF", "FS", "FSF"}, the standard "SF" way is used.
    """
    
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    
    # Filtering
    if type(spec) is CEX:
        winning = cex_si(fsm, agents,
                         evalATLK(fsm, spec.child, variant=variant), strat)

    elif type(spec) is CEG:
        winning = ceg_si(fsm, agents, evalATLK(fsm, spec.child, variant=variant), strat)

    elif type(spec) is CEU:
        winning = ceu_si(fsm, agents, evalATLK(fsm, spec.left, variant=variant),
                         evalATLK(fsm, spec.right, variant=variant), strat)

    elif type(spec) is CEF:
        # <g> F p = <g>[true U p]
        winning = ceu_si(fsm, agents, BDD.true(fsm.bddEnc.DDmanager),
                         evalATLK(fsm, spec.child, variant=variant), strat)

    elif type(spec) is CEW:
       winning = cew_si(fsm, agents, evalATLK(fsm, spec.left, variant=variant),
                        evalATLK(fsm, spec.right, variant=variant), strat)
    
    
    return winning & fsm.bddEnc.statesInputsMask & fsm.protocol(agents)
    
    
def eval_strat_FSF(fsm, spec):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    spec is a strategic operator <G> pi.
    
    Implement a variant of the algorithm that
    filters, splits and then filters.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator.
    
    """
    
    if type(spec) is CAX:
        # [g] X p = ~<g> X ~p
        newspec = CEX(spec.group, Not(spec.child))
        return ~eval_strat_FSF(fsm, newspec)
        
    elif type(spec) is CAG:
        # [g] G p = ~<g> F ~p
        newspec = CEF(spec.group, Not(spec.child))
        return ~eval_strat_FSF(fsm, newspec)
        
    elif type(spec) is CAU:
        # [g][p U q] = ~<g>[ ~q W ~p & ~q ]
        newspec = CEW(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~eval_strat_FSF(fsm, newspec)
        
    elif type(spec) is CAF:
        # [g] F p = ~<g> G ~p
        newspec = CEG(spec.group, Not(spec.child))
        return ~eval_strat_FSF(fsm, newspec)
        
    elif type(spec) is CAW:
        # [g][p W q] = ~<g>[~q U ~p & ~q]
        newspec = CEU(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~eval_strat_FSF(fsm, newspec)
        
        
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    
    # First filtering
    winning = filter_strat(fsm, spec, variant="FSF")
    
    if winning.is_false(): # no state/inputs pairs are winning => return false
        return winning
    
    return split_eval(fsm, spec, BDD.false(fsm.bddEnc.DDmanager),
                      winning)


"""
The objective of the tool implemented in :mod:`pynusmv_tools.bmcLTL` is to 
demonstrate with a simple example (LTL) how one can use the bmc additions 
to PyNuSMV to develop verification tools for new logic formalisms. Hence, 
the approach that was adopted in that context was to *NOT* use any of the 
"higher level services" and develop a verification tool translating 
"literally" the reduction of that logic to a propositional SAT problem as 
imagined or stated in a reference paper.

In this scope, the reference paper that was used was:

            Biere et al - ``Bounded Model Checking'' - 2003  
"""
__all__ = ['parsing', 'ast', 'gen', 'check']


from ..eval import evalCTLK
from .explain import explain_witness, explain_countex

def checkCTLK(fsm, spec):
    """
    Check whether fsm satisfies spec.
    
    fsm -- a MAS
    spec -- a CTLK specification
    
    Return a tuple (sat, diag) where
        sat is True iff fsm satisfies spec, it is False otherwise;
        diag is a partial TLACE explaining why fsm satisfies or violates spec.
    """
    
    sat = evalCTLK(fsm, spec)
    violating = fsm.init & ~sat
    if violating.isnot_false():
        return (False,
                explain_countex(fsm, fsm.pick_one_state(violating), spec))
    else:
        return (True,
                explain_witness(fsm, fsm.pick_one_state(fsm.init), spec))


import unittest

from pynusmv_lower_interface.nusmv.cmd import cmd
from pynusmv_lower_interface.nusmv.compile.symb_table import symb_table
from pynusmv_lower_interface.nusmv.enc.bdd import bdd as bddEnc

from pynusmv.fsm import BddFsm
from pynusmv.prop import PropDb
from pynusmv import glob

from pynusmv_tools.tlace.check import check
from pynusmv_tools.tlace.xml import xml_representation

from pynusmv.init import init_nusmv, deinit_nusmv

class TestTlacePrinting(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
        
    def tearDown(self):
        deinit_nusmv()
    
    
    def test_print_violated_spec_admin(self):
        # Initialize the model
        ret = cmd.Cmd_SecureCommandExecute("read_model -i " 
                                     "tests/pynusmv_tools/tlace/admin.smv")
        self.assertEqual(ret, 0, "cannot read the model")
        ret = cmd.Cmd_SecureCommandExecute("go")
        self.assertEqual(ret, 0, "cannot build the model")
        
        propDb = glob.prop_database()
        master = propDb.master
        fsm = propDb.master.bddFsm
        self.assertTrue(propDb.get_size() >= 1, "propDb has no properties")
        prop = propDb.get_prop_at_index(0)
        spec = prop.exprcore
        
        res = check(fsm, spec)
        self.assertFalse(res[0], "spec should be violated")
        self.assertIsNotNone(res[1], "TLACE should be given")
        
        print(xml_representation(fsm, res[1], spec))
        
        
    def test_print_violated_spec_egmod(self):
        # Initialize the model
        ret = cmd.Cmd_SecureCommandExecute("read_model -i " 
                                     "tests/pynusmv_tools/tlace/eg.smv")
        self.assertEqual(ret, 0, "cannot read the model")
        ret = cmd.Cmd_SecureCommandExecute("go")
        self.assertEqual(ret, 0, "cannot build the model")
        
        propDb = glob.prop_database()
        master = propDb.master
        fsm = propDb.master.bddFsm
        self.assertTrue(propDb.get_size() >= 1, "propDb has no properties")
        prop = propDb.get_prop_at_index(0)
        spec = prop.exprcore
        
        res = check(fsm, spec)
        self.assertFalse(res[0], "spec should be violated")
        self.assertIsNotNone(res[1], "TLACE should be given")

        print(xml_representation(fsm, res[1], spec))
        
    
    def test_print_violated_spec_egmod2(self):
        # Initialize the model
        ret = cmd.Cmd_SecureCommandExecute("read_model -i " 
                                     "tests/pynusmv_tools/tlace/eg.smv")
        self.assertEqual(ret, 0, "cannot read the model")
        ret = cmd.Cmd_SecureCommandExecute("go")
        self.assertEqual(ret, 0, "cannot build the model")
        
        propDb = glob.prop_database()
        master = propDb.master
        fsm = propDb.master.bddFsm
        self.assertTrue(propDb.get_size() >= 2, "propDb has no properties")
        prop = propDb.get_prop_at_index(1)
        spec = prop.exprcore
        
        res = check(fsm, spec)
        self.assertFalse(res[0], "spec should be violated")
        self.assertIsNotNone(res[1], "TLACE should be given")

        print(xml_representation(fsm, res[1], spec))
        
    
    def test_print_violated_spec_admin_af(self):
        # Initialize the model
        ret = cmd.Cmd_SecureCommandExecute("read_model -i " 
                                     "tests/pynusmv_tools/tlace/admin.smv")
        self.assertEqual(ret, 0, "cannot read the model")
        ret = cmd.Cmd_SecureCommandExecute("go")
        self.assertEqual(ret, 0, "cannot build the model")
        
        propDb = glob.prop_database()
        master = propDb.master
        fsm = propDb.master.bddFsm
        self.assertTrue(propDb.get_size() >= 5, "propDb has no properties")
        prop = propDb.get_prop_at_index(4)
        spec = prop.exprcore
        
        res = check(fsm, spec)
        self.assertFalse(res[0], "spec should be violated")
        self.assertIsNotNone(res[1], "TLACE should be given")

        print(xml_representation(fsm, res[1], spec))
        
        
    def test_print_violated_spec_admin_ax(self):
        # Initialize the model
        ret = cmd.Cmd_SecureCommandExecute("read_model -i " 
                                     "tests/pynusmv_tools/tlace/admin.smv")
        self.assertEqual(ret, 0, "cannot read the model")
        ret = cmd.Cmd_SecureCommandExecute("go")
        self.assertEqual(ret, 0, "cannot build the model")
        
        propDb = glob.prop_database()
        master = propDb.master
        fsm = propDb.master.bddFsm
        self.assertTrue(len(propDb) >= 6, "propDb has no properties")
        prop = propDb[5]
        spec = prop.exprcore
        
        res = check(fsm, spec)
        self.assertFalse(res[0], "spec should be violated")
        self.assertIsNotNone(res[1], "TLACE should be given")

        print(xml_representation(fsm, res[1], spec))


import unittest

from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.mc import eval_simple_expression

from pynusmv_tools.mas import glob
from pynusmv_tools.ctlk.ast import (TrueExp, FalseExp, Init, Reachable,
                            Atom, Not, And, Or, Implies, Iff, 
                            AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                            nK, nE, nD, nC, K, E, D, C)
from pynusmv_tools.ctlk.parsing import parseCTLK
from pynusmv_tools.ctlk.eval import evalCTLK
from pynusmv_tools.ctlk.tlace.tlace import Tlacenode, TemporalBranch, EpistemicBranch
from pynusmv_tools.ctlk.tlace.explain import explain_witness, explain_countex
from pynusmv_tools.ctlk.tlace.xml import xml_countex, xml_witness

class TestXML(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
        
    def tearDown(self):
        glob.reset_globals()
        deinit_nusmv()
        
    def model(self):
        glob.load_from_file("tests/pynusmv_tools/ctlk/dining-crypto.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
        
    def simplemodel(self):
        glob.load_from_file("tests/pynusmv_tools/ctlk/agents.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
        
    def test_true(self):
        fsm = self.simplemodel()
        
        spec = parseCTLK("True")[0]
        self.assertIsNotNone(spec)
        state = fsm.pick_one_state(fsm.init & evalCTLK(fsm, spec))
        self.assertIsNotNone(state)
        
        expl = explain_witness(fsm, state, spec)
        self.assertIsNotNone(expl)
        self.assertEqual(type(expl), Tlacenode)
        self.assertEqual(expl.state, state)
        self.assertEqual(len(expl.atomics), 0)
        self.assertEqual(len(expl.branches), 0)
        self.assertEqual(len(expl.universals), 0)
        
        print(xml_witness(fsm, expl, spec))
        
        
    def test_false(self):
        fsm = self.simplemodel()
        
        spec = parseCTLK("False")[0]
        self.assertIsNotNone(spec)
        state = fsm.pick_one_state(fsm.init & ~evalCTLK(fsm, spec))
        self.assertIsNotNone(state)
        
        expl = explain_countex(fsm, state, spec)
        self.assertIsNotNone(expl)
        self.assertEqual(type(expl), Tlacenode)
        self.assertEqual(expl.state, state)
        self.assertEqual(len(expl.atomics), 0)
        self.assertEqual(len(expl.branches), 0)
        self.assertEqual(len(expl.universals), 0)
        
        print(xml_countex(fsm, expl, spec))
        
        
    def test_init(self):
        fsm = self.simplemodel()
        
        spec = parseCTLK("Init")[0]
        self.assertIsNotNone(spec)
        state = fsm.pick_one_state(fsm.init & evalCTLK(fsm, spec))
        self.assertIsNotNone(state)
        
        expl = explain_witness(fsm, state, spec)
        self.assertIsNotNone(expl)
        self.assertEqual(type(expl), Tlacenode)
        self.assertEqual(expl.state, state)
        self.assertEqual(len(expl.atomics), 1)
        self.assertEqual(type(expl.atomics[0]), Init)
        self.assertEqual(len(expl.branches), 0)
        self.assertEqual(len(expl.universals), 0)
        
        print(xml_witness(fsm, expl, spec))
        
        
    def test_reachable(self):
        fsm = self.simplemodel()
        
        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        spec = parseCTLK("Reachable")[0]
        self.assertIsNotNone(spec)
        state = fsm.pick_one_state(~lt & ~lf & ~g)
        self.assertIsNotNone(state)
        self.assertTrue(state <= evalCTLK(fsm, spec))
        
        expl = explain_witness(fsm, state, spec)
        self.assertIsNotNone(expl)
        self.assertEqual(type(expl), Tlacenode)
        self.assertEqual(expl.state, state)
        self.assertEqual(len(expl.atomics), 0)
        self.assertEqual(len(expl.branches), 1)
        self.assertEqual(len(expl.universals), 0)
        
        branch = expl.branches[0]
        self.assertEqual(type(branch), TemporalBranch)
        self.assertEqual(type(branch.specification), Reachable)
        path = branch.path
        self.assertIsNone(branch.loop)
        for (s, i, sp) in zip(path[::2], path[1::2], path[2::2]):
            self.assertIsNotNone(s)
            self.assertTrue(s.state <= fsm.reachable_states)
            self.assertEqual(len(s.atomics), 0)
            self.assertEqual(len(s.branches), 0)
            self.assertEqual(len(s.universals), 0)
            self.assertIsNotNone(sp)
            self.assertTrue(sp.state <= fsm.reachable_states)
            self.assertEqual(len(sp.branches), 0)
            self.assertEqual(len(sp.universals), 0)
            self.assertTrue(i <=
                               fsm.get_inputs_between_states(sp.state, s.state))
        self.assertEqual(len(path[-1].atomics), 1)
        self.assertEqual(type(path[-1].atomics[0]), Init)
        
        print(xml_witness(fsm, expl, spec))
        
        
    def test_full_witness(self):
        fsm = self.model()
        
        spec = parseCTLK("EF (EX nK<'c1'> 'c2.payer' & EG nD<'c1','c2'> 'c3.payer')")[0]
        self.assertIsNotNone(spec)
        state = fsm.pick_one_state(fsm.init & evalCTLK(fsm, spec))
        self.assertIsNotNone(state)
        
        expl = explain_witness(fsm, state, spec)
        self.assertIsNotNone(expl)
        
        print(xml_witness(fsm, expl, spec))
        
        
        spec = parseCTLK("E[nE<'c1','c2'> 'c3.payer' U"
                        " E[ nC<'c1','c3'> ('c1.payer' | 'c2.payer') W"
                        " nC<'c2'> ('c1.coin = head' <-> 'c2.coin = head')]]")[0]
        self.assertIsNotNone(spec)
        state = fsm.pick_one_state(fsm.init & evalCTLK(fsm, spec))
        self.assertIsNotNone(state)
        
        expl = explain_witness(fsm, state, spec)
        self.assertIsNotNone(expl)
        
        print(xml_witness(fsm, expl, spec))
        
        
        spec = parseCTLK("AG(~'c1.payer' | E<'c1'> ('c1.payer' -> "
                         "C<'c2','c3'> AF 'c1.payer'))")[0]
        self.assertIsNotNone(spec)
        state = fsm.pick_one_state(fsm.init & evalCTLK(fsm, spec))
        self.assertIsNotNone(state)
        
        expl = explain_witness(fsm, state, spec)
        self.assertIsNotNone(expl)
        
        print(xml_witness(fsm, expl, spec))
        
        
    def test_full_countex(self):
        fsm = self.model()
        
        spec = parseCTLK("'c2.payer' -> A[AX ('c1.payer' | ~'c1.payer')"
                         " U K<'c1'> ('c2.payer')]")[0]
        self.assertIsNotNone(spec)
        self.assertTrue((fsm.init & ~evalCTLK(fsm, spec)).isnot_false())
        state = fsm.pick_one_state(fsm.init & ~evalCTLK(fsm, spec))
        self.assertIsNotNone(state)
        
        expl = explain_countex(fsm, state, spec)
        self.assertIsNotNone(expl)
        
        print(xml_countex(fsm, expl, spec))
        
        
        spec = parseCTLK("AG(~'c1.payer' | E<'c1'> ('c1.payer' -> "
                         "C<'c2','c3'> AF 'c1.payer'))")[0]
        self.assertIsNotNone(spec)
        self.assertTrue((fsm.init & ~evalCTLK(fsm, spec)).isnot_false())
        state = fsm.pick_one_state(fsm.init & ~evalCTLK(fsm, spec))
        self.assertIsNotNone(state)
        
        expl = explain_countex(fsm, state, spec)
        self.assertIsNotNone(expl)
        
        print(xml_countex(fsm, expl, spec))